"""
Conversion utilities for mapping memory, disk, core declarations from strings to numbers and vice versa.
Also contains general conversion functions
"""

import math
from typing import Optional, SupportsInt, Union

# See https://en.wikipedia.org/wiki/Binary_prefix
BINARY_PREFIXES = [
    "ki",
    "mi",
    "gi",
    "ti",
    "pi",
    "ei",
    "kib",
    "mib",
    "gib",
    "tib",
    "pib",
    "eib",
]
DECIMAL_PREFIXES = [
    "b",
    "k",
    "m",
    "g",
    "t",
    "p",
    "e",
    "kb",
    "mb",
    "gb",
    "tb",
    "pb",
    "eb",
]
VALID_PREFIXES = BINARY_PREFIXES + DECIMAL_PREFIXES


def bytes_in_unit(unit: str = "B") -> int:
    num_bytes = 1
    if unit.lower() in ["ki", "kib"]:
        num_bytes = 1 << 10
    if unit.lower() in ["mi", "mib"]:
        num_bytes = 1 << 20
    if unit.lower() in ["gi", "gib"]:
        num_bytes = 1 << 30
    if unit.lower() in ["ti", "tib"]:
        num_bytes = 1 << 40
    if unit.lower() in ["pi", "pib"]:
        num_bytes = 1 << 50
    if unit.lower() in ["ei", "eib"]:
        num_bytes = 1 << 60

    if unit.lower() in ["k", "kb"]:
        num_bytes = 1000
    if unit.lower() in ["m", "mb"]:
        num_bytes = 1000**2
    if unit.lower() in ["g", "gb"]:
        num_bytes = 1000**3
    if unit.lower() in ["t", "tb"]:
        num_bytes = 1000**4
    if unit.lower() in ["p", "pb"]:
        num_bytes = 1000**5
    if unit.lower() in ["e", "eb"]:
        num_bytes = 1000**6
    return num_bytes


def convert_units(num: float, src_unit: str, dst_unit: str = "B") -> float:
    """Returns a float representing the converted input in dst_units."""
    if not src_unit.lower() in VALID_PREFIXES:
        raise RuntimeError(
            f"{src_unit} not a valid unit, valid units are {VALID_PREFIXES}."
        )
    if not dst_unit.lower() in VALID_PREFIXES:
        raise RuntimeError(
            f"{dst_unit} not a valid unit, valid units are {VALID_PREFIXES}."
        )
    return (num * bytes_in_unit(src_unit)) / bytes_in_unit(dst_unit)


def parse_memory_string(string: str) -> tuple[float, str]:
    """
    Given a string representation of some memory (i.e. '1024 Mib'), return the
    number and unit.
    """
    for i, character in enumerate(string):
        # find the first character of the unit
        if character not in "0123456789.-_ ":
            units = string[i:].strip()
            if not units.lower() in VALID_PREFIXES:
                raise RuntimeError(
                    f"{units} not a valid unit, valid units are {VALID_PREFIXES}."
                )
            return float(string[:i]), units
    return float(string), "b"


def human2bytes(string: str) -> int:
    """
    Given a string representation of some memory (i.e. '1024 Mib'), return the
    integer number of bytes.
    """
    value, unit = parse_memory_string(string)

    return int(convert_units(value, src_unit=unit, dst_unit="b"))


def bytes2human(n: SupportsInt) -> str:
    """Return a binary value as a human readable string with units."""
    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    elif n < 1:
        return "0 b"

    power_level = math.floor(math.log(n, 1024))
    units = ("b", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei")

    unit = units[power_level if power_level < len(units) else -1]
    value = convert_units(n, "b", unit)
    return f"{value:.1f} {unit}"


def b_to_mib(n: Union[int, float]) -> float:
    """
    Convert a number from bytes to mibibytes.
    """
    return convert_units(n, "b", "mib")


def mib_to_b(n: Union[int, float]) -> float:
    """
    Convert a number from mibibytes to bytes.
    """
    return convert_units(n, "mib", "b")


# General Conversions


def hms_duration_to_seconds(hms: str) -> float:
    """
    Parses a given time string in hours:minutes:seconds,
    returns an equivalent total seconds value
    """
    vals_to_convert = hms.split(":")
    seconds = 0.0

    for val in vals_to_convert:
        if float(val) < 0:
            raise ValueError("Invalid Time, negative value")

    if len(vals_to_convert) != 3:
        raise ValueError("Invalid amount of fields, function takes input in 'hh:mm:ss'")

    seconds += float(vals_to_convert[0]) * 60 * 60
    seconds += float(vals_to_convert[1]) * 60
    seconds += float(vals_to_convert[2])

    return seconds


def strtobool(val: str) -> bool:
    """
    Make a human-readable string into a bool.

    Convert a string along the lines of "y", "1", "ON", "TrUe", or
    "Yes" to True, and the corresponding false-ish values to False.
    """
    # We only track prefixes, so "y" covers "y", "yes",
    # and "yeah no" and makes them all True.
    TABLE = {True: ["1", "on", "y", "t"], False: ["0", "off", "n", "f"]}
    lowered = val.lower()
    for result, prefixes in TABLE.items():
        for prefix in prefixes:
            if lowered.startswith(prefix):
                return result
    raise ValueError(f'Cannot convert "{val}" to a bool')


def opt_strtobool(b: Optional[str]) -> Optional[bool]:
    """Convert an optional string representation of bool to None or bool"""
    return b if b is None else strtobool(b)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5.14.2018: copied into Toil from https://github.com/BD2KGenomics/bd2k-python-lib
# Note: renamed from "threading.py" to "threading.py" to avoid conflicting imports
# from the built-in "threading" from psutil in python3.9
import atexit
import errno
import fcntl
import logging
import math
import os
import platform
import subprocess
import sys
import tempfile
import threading
import time
import traceback
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Optional, Union, cast

import psutil

from toil.lib.exceptions import raise_
from toil.lib.io import robust_rmtree
from toil.lib.misc import StrPath

logger = logging.getLogger(__name__)


def ensure_filesystem_lockable(
    path: StrPath, timeout: float = 30, hint: Optional[str] = None
) -> None:
    """
    Make sure that the filesystem used at the given path is one where locks are safe to use.

    File locks are not safe to use on Ceph. See
    <https://github.com/DataBiosphere/toil/issues/4972>.

    Raises an exception if the filesystem is detected as one where using locks
    is known to trigger bugs in the filesystem implementation. Also raises an
    exception if the given path does not exist, or if attempting to determine
    the filesystem type takes more than the timeout in seconds.

    If the filesystem type cannot be determined, does nothing.

    :param hint: Extra text to include in an error, if raised, telling the user
        how to change the offending path.
    """

    if not os.path.exists(path):
        # Raise a normal-looking FileNotFoundError. See <https://stackoverflow.com/a/36077407>
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    if platform.system() == "Linux":
        # We know how to find the filesystem here.

        try:
            # Start a child process to stat the path. See <https://unix.stackexchange.com/a/402236>.
            # We really should call statfs but no bindings for it are in PyPI.
            completed = subprocess.run(
                ["stat", "-f", "-c", "%T", str(path)],
                check=True,
                capture_output=True,
                timeout=timeout,
            )
        except subprocess.TimeoutExpired as e:
            # The subprocess itself is Too Slow
            raise RuntimeError(
                f"Polling filesystem type at {path} took more than {timeout} seconds; is your filesystem working?"
            ) from e
        except subprocess.CalledProcessError as e:
            # Stat didn't work. Maybe we don't have the right version of stat installed?
            logger.warning(
                "Could not determine filesystem type at %s because of: %s",
                str(path),
                e.stderr.decode("utf-8", errors="replace").strip(),
            )
            # If we don't know the filesystem type, keep going anyway.
            return

        filesystem_type = completed.stdout.decode("utf-8", errors="replace").strip()

        if filesystem_type == "ceph":
            # Ceph is known to deadlock the MDS and break the parent directory when locking.
            message = [
                f"Refusing to use {path} because file locks are known to break {filesystem_type} filesystems."
            ]
            if hint:
                # Hint the user how to fix this.
                message.append(hint)
            raise RuntimeError(" ".join(message))
        else:
            # Other filesystem types are fine (even though NFS is sometimes
            # flaky with regard to locks actually locking anything).
            logger.debug(
                "Detected that %s has lockable filesystem type: %s",
                str(path),
                filesystem_type,
            )

    # Other platforms (Mac) probably aren't mounting Ceph and also don't
    # usually use the same stat binary implementation.


def safe_lock(fd: int, block: bool = True, shared: bool = False) -> None:
    """
    Get an fcntl lock, while retrying on IO errors.

    Raises OSError with EACCES or EAGAIN when a nonblocking lock is not
    immediately available.
    """

    # Set up retry logic. TODO: Use @retry instead.
    error_backoff = 1
    MAX_ERROR_TRIES = 10
    error_tries = 0

    while True:
        try:
            # Wait until we can exclusively lock it.
            lock_mode = (fcntl.LOCK_SH if shared else fcntl.LOCK_EX) | (
                fcntl.LOCK_NB if not block else 0
            )
            fcntl.flock(fd, lock_mode)
            return
        except OSError as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                # Nonblocking lock not available.
                raise
            elif e.errno == errno.EIO:
                # Sometimes Ceph produces IO errors when talking to lock files.
                # Back off and try again.
                # TODO: Should we eventually give up if the disk really is
                # broken? If so we should use the retry system.
                if error_tries < MAX_ERROR_TRIES:
                    logger.error(
                        "IO error talking to lock file. Retrying after %s seconds.",
                        error_backoff,
                    )
                    time.sleep(error_backoff)
                    error_backoff = min(60, error_backoff * 2)
                    error_tries += 1
                    continue
                else:
                    logger.critical(
                        "Too many IO errors talking to lock file. If using Ceph, check for MDS deadlocks. See <https://tracker.ceph.com/issues/62123>."
                    )
                    raise
            else:
                raise


def safe_unlock_and_close(fd: int) -> None:
    """
    Release an fcntl lock and close the file descriptor, while handling fcntl IO errors.
    """
    try:
        fcntl.flock(fd, fcntl.LOCK_UN)
    except OSError as e:
        if e.errno != errno.EIO:
            raise
        # Sometimes Ceph produces EIO. We don't need to retry then because
        # we're going to close the FD and after that the file can't remain
        # locked by us.
    os.close(fd)


class ExceptionalThread(threading.Thread):
    """
    A thread whose join() method re-raises exceptions raised during run(). While join() is
    idempotent, the exception is only during the first invocation of join() that successfully
    joined the thread. If join() times out, no exception will be re reraised even though an
    exception might already have occurred in run().

    When subclassing this thread, override tryRun() instead of run().

    >>> def f():
    ...     assert 0
    >>> t = ExceptionalThread(target=f)
    >>> t.start()
    >>> t.join()
    Traceback (most recent call last):
    ...
    AssertionError

    >>> class MyThread(ExceptionalThread):
    ...     def tryRun( self ):
    ...         assert 0
    >>> t = MyThread()
    >>> t.start()
    >>> t.join()
    Traceback (most recent call last):
    ...
    AssertionError

    """

    exc_info = None

    def run(self) -> None:
        try:
            self.tryRun()
        except:
            self.exc_info = sys.exc_info()
            raise

    def tryRun(self) -> None:
        super().run()

    def join(self, *args: Optional[float], **kwargs: Optional[float]) -> None:
        super().join(*args, **kwargs)
        if not self.is_alive() and self.exc_info is not None:
            exc_type, exc_value, traceback = self.exc_info
            self.exc_info = None
            raise_(exc_type, exc_value, traceback)


def cpu_count() -> int:
    """
    Get the rounded-up integer number of whole CPUs available.

    Counts hyperthreads as CPUs.

    Uses the system's actual CPU count, or the current v1 cgroup's quota per
    period, if the quota is set.

    Ignores the cgroup's cpu shares value, because it's extremely difficult to
    interpret. See https://github.com/kubernetes/kubernetes/issues/81021.

    Caches result for efficiency.

    :return: Integer count of available CPUs, minimum 1.
    :rtype: int
    """

    cached = getattr(cpu_count, "result", None)
    if cached is not None:
        # We already got a CPU count.
        return cast(int, cached)

    # Get the fallback answer of all the CPUs on the machine
    psutil_cpu_count = psutil.cpu_count(logical=True)
    if psutil_cpu_count is None:
        logger.debug("Could not retrieve the logical CPU count.")

    total_machine_size: Union[float, int] = (
        psutil_cpu_count if psutil_cpu_count is not None else float("inf")
    )
    logger.debug("Total machine size: %s core(s)", total_machine_size)

    # cgroups may limit the size
    cgroup_size: Union[float, int] = float("inf")

    try:
        # See if we can fetch these and use them
        quota: Optional[int] = None
        period: Optional[int] = None

        # CGroups v1 keeps quota and period separate
        CGROUP1_QUOTA_FILE = "/sys/fs/cgroup/cpu/cpu.cfs_quota_us"
        CGROUP1_PERIOD_FILE = "/sys/fs/cgroup/cpu/cpu.cfs_period_us"
        # CGroups v2 keeps both in one file, space-separated, quota first
        CGROUP2_COMBINED_FILE = "/sys/fs/cgroup/cpu.max"

        if os.path.exists(CGROUP1_QUOTA_FILE) and os.path.exists(CGROUP1_PERIOD_FILE):
            logger.debug("CPU quota and period available from cgroups v1")
            with open(CGROUP1_QUOTA_FILE) as stream:
                # Read the quota
                quota = int(stream.read())

            with open(CGROUP1_PERIOD_FILE) as stream:
                # Read the period in which we are allowed to burn the quota
                period = int(stream.read())
        elif os.path.exists(CGROUP2_COMBINED_FILE):
            logger.debug("CPU quota and period available from cgroups v2")
            with open(CGROUP2_COMBINED_FILE) as stream:
                # Read the quota and the period together
                quota, period = (int(part) for part in stream.read().split(" "))
        else:
            logger.debug("CPU quota/period not available from cgroups v1 or cgroups v2")

        if quota is not None and period is not None:
            # We got a quota and a period.
            logger.debug("CPU quota: %d period: %d", quota, period)

            if quota == -1:
                # But the quota can be -1 for unset.
                # Assume we can use the whole machine.
                cgroup_size = float("inf")
            else:
                # The thread count is how many multiples of a wall clock period we
                # can burn in that period.
                cgroup_size = int(math.ceil(float(quota) / float(period)))

            logger.debug("Control group size in cores: %s", cgroup_size)
    except:
        # We can't actually read these cgroup fields. Maybe we are a mac or something.
        logger.debug("Could not inspect cgroup: %s", traceback.format_exc())

    # CPU affinity may limit the size
    affinity_size: Union[float, int] = float("inf")
    if hasattr(os, "sched_getaffinity"):
        try:
            logger.debug("CPU affinity available")
            affinity_size = len(os.sched_getaffinity(0))
            logger.debug("CPU affinity is restricted to %d cores", affinity_size)
        except:
            # We can't actually read this even though it exists.
            logger.debug(
                "Could not inspect scheduling affinity: %s", traceback.format_exc()
            )
    else:
        logger.debug("CPU affinity not available")

    limit: Union[float, int] = float("inf")
    # Apply all the limits to take the smallest
    limit = min(limit, total_machine_size)
    limit = min(limit, cgroup_size)
    limit = min(limit, affinity_size)
    if limit < 1 or limit == float("inf"):
        # Fall back to 1 if we can't get a size
        limit = 1
    result = int(limit)
    logger.debug("cpu_count: %s", result)
    # Make sure to remember it for the next call
    setattr(cpu_count, "result", result)
    return result


# PIDs are a bad identifier, because they are not shared between containers
# and also may be reused.
# So instead we have another system for file store implementations to
# coordinate among themselves, based on file locks.
# TODO: deduplicate with DeferredFunctionManager?
# TODO: Wrap in a class as static methods?

# Note that we don't offer a way to enumerate these names. You can only get
# your name and poll others' names (or your own). So we don't have
# distinguishing prefixes or WIP suffixes to allow for enumeration.

# We keep one name per unique base directory (probably a Toil coordination
# directory).

# We have a global lock to control looking things up
current_process_name_lock = threading.Lock()
# And a global dict from work directory to name in that work directory.
# We also have a file descriptor per work directory but it is just leaked.
current_process_name_for: dict[str, str] = {}


def collect_process_name_garbage() -> None:
    """
    Delete all the process names that point to files that don't exist anymore
    (because the work directory was temporary and got cleaned up). This is
    known to happen during the tests, which get their own temp directories.

    Caller must hold current_process_name_lock.
    """

    global current_process_name_for

    # Collect the base_dirs of the missing names to delete them after iterating.
    missing = []

    for base_dir, name in current_process_name_for.items():
        if not os.path.exists(os.path.join(base_dir, name)):
            # The name file is gone, probably because the work dir is gone.
            missing.append(base_dir)

    for base_dir in missing:
        del current_process_name_for[base_dir]


def destroy_all_process_names() -> None:
    """
    Delete all our process name files because our process is going away.

    We let all our FDs get closed by the process death.

    We assume there is nobody else using the system during exit to race with.
    """

    global current_process_name_for

    for base_dir, name in current_process_name_for.items():
        robust_rmtree(os.path.join(base_dir, name))


# Run the cleanup at exit
atexit.register(destroy_all_process_names)


def get_process_name(base_dir: str) -> str:
    """
    Return the name of the current process. Like a PID but visible between
    containers on what to Toil appears to be a node.

    :param str base_dir: Base directory to work in. Defines the shared namespace.
    :return: Process's assigned name
    :rtype: str
    """

    global current_process_name_lock
    global current_process_name_for

    with current_process_name_lock:

        # Make sure all the names still exist.
        # TODO: a bit O(n^2) in the number of base_dirs in flight at any one time.
        collect_process_name_garbage()

        if base_dir in current_process_name_for:
            # If we already gave ourselves a name, return that.
            return current_process_name_for[base_dir]

        # We need to get a name file.
        nameFD, nameFileName = tempfile.mkstemp(dir=base_dir)

        # Lock the file. The lock will automatically go away if our process does.
        try:
            safe_lock(nameFD, block=False)
        except OSError as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                # Someone else locked it even though they should not have.
                raise RuntimeError(
                    f"Could not lock process name file {nameFileName}"
                ) from e
            else:
                # Something else is wrong
                raise

        # Save the basename
        current_process_name_for[base_dir] = os.path.basename(nameFileName)

        # Return the basename
        return current_process_name_for[base_dir]

        # TODO: we leave the file open forever. We might need that in order for
        # it to stay locked while we are alive.


def process_name_exists(base_dir: str, name: str) -> bool:
    """
    Return true if the process named by the given name (from process_name) exists, and false otherwise.

    Can see across container boundaries using the given node workflow directory.

    :param str base_dir: Base directory to work in. Defines the shared namespace.
    :param str name: Process's name to poll
    :return: True if the named process is still alive, and False otherwise.
    :rtype: bool
    """

    global current_process_name_lock
    global current_process_name_for

    with current_process_name_lock:
        if current_process_name_for.get(base_dir, None) == name:
            # We are asking about ourselves. We are alive.
            return True

    # Work out what the corresponding file name is
    nameFileName = os.path.join(base_dir, name)
    if not os.path.exists(nameFileName):
        # If the file is gone, the process can't exist.
        return False

    nameFD = None
    try:
        try:
            # Otherwise see if we can lock it shared, for which we need an FD, but
            # only for reading.
            nameFD = os.open(nameFileName, os.O_RDONLY)
        except FileNotFoundError as e:
            # File has vanished
            return False
        try:
            safe_lock(nameFD, block=False, shared=True)
        except OSError as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                # Could not lock. Process is alive.
                return True
            else:
                # Something else went wrong
                raise
        else:
            # Could lock. Process is dead.
            # Remove the file. We race to be the first to do so.
            try:
                os.remove(nameFileName)
            except:
                pass
            safe_unlock_and_close(nameFD)
            nameFD = None
            # Report process death
            return False
    finally:
        if nameFD is not None:
            try:
                os.close(nameFD)
            except:
                pass


# Similar to the process naming system above, we define a global mutex system
# for critical sections, based just around file locks.
@contextmanager
def global_mutex(base_dir: StrPath, mutex: str) -> Iterator[None]:
    """
    Context manager that locks a mutex. The mutex is identified by the given
    name, and scoped to the given directory. Works across all containers that
    have access to the given diectory. Mutexes held by dead processes are
    automatically released.

    Only works between processes, NOT between threads.

    :param base_dir: Base directory to work in. Defines the shared namespace.
    :param str mutex: Mutex to lock. Must be a permissible path component.
    """

    if not os.path.isdir(base_dir):
        raise RuntimeError(f"Directory {base_dir} for mutex does not exist")

    # TODO: We don't know what CLI option controls where to put this mutex, so
    # we aren't very helpful if the location is bad.
    ensure_filesystem_lockable(
        base_dir, hint=f"Specify a different place to put the {mutex} mutex."
    )

    # Define a filename
    lock_filename = os.path.join(base_dir, "toil-mutex-" + mutex)

    logger.debug("PID %d acquiring mutex %s", os.getpid(), lock_filename)

    # We can't just create/open and lock a file, because when we clean up
    # there's a race where someone can open the file before we unlink it and
    # get a lock on the deleted file.

    error_backoff = 1

    while True:
        # Try to create the file, ignoring if it exists or not.
        fd = os.open(lock_filename, os.O_CREAT | os.O_WRONLY)

        try:
            # Wait until we can exclusively lock it, handling error retry.
            safe_lock(fd)
        except:
            # Something went wrong
            os.close(fd)
            raise

        # Holding the lock, make sure we are looking at the same file on disk still.
        try:
            # So get the stats from the open file
            fd_stats = os.fstat(fd)
        except OSError as e:
            if e.errno == errno.ESTALE:
                # The file handle has gone stale, because somebody removed the
                # file.
                # Try again.
                safe_unlock_and_close(fd)
                continue
            else:
                # Something else broke
                os.close(fd)
                raise

        try:
            # And get the stats for the name in the directory
            path_stats: Optional[os.stat_result] = os.stat(lock_filename)
        except FileNotFoundError:
            path_stats = None

        if (
            path_stats is None
            or fd_stats.st_dev != path_stats.st_dev
            or fd_stats.st_ino != path_stats.st_ino
        ):
            # The file we have a lock on is not the file linked to the name (if
            # any). This usually happens, because before someone releases a
            # lock, they delete the file. Go back and contend again. TODO: This
            # allows a lot of queue jumping on our mutex.
            safe_unlock_and_close(fd)
            continue
        else:
            # We have a lock on the file that the name points to. Since we
            # hold the lock, nobody will be deleting it or can be in the
            # process of deleting it. Stop contending; we have the mutex.
            break

    try:
        # When we have it, do the thing we are protecting.
        logger.debug("PID %d now holds mutex %s", os.getpid(), lock_filename)
        yield
    finally:
        # Delete it while we still own it, so we can't delete it from out from
        # under someone else who thinks they are holding it.
        logger.debug("PID %d releasing mutex %s", os.getpid(), lock_filename)

        # We have had observations in the wild of the lock file not exisiting
        # when we go to unlink it, causing a crash on mutex release. See
        # <https://github.com/DataBiosphere/toil/issues/4654>.
        #
        # We want to tolerate this; maybe unlink() interacts with fcntl() locks
        # on NFS in a way that is actually fine, somehow? But we also want to
        # complain loudly if something is tampering with our locks or not
        # really enforcing locks on the filesystem, so we will notice if it is
        # the cause of further problems.
        try:
            path_stats = os.stat(lock_filename)
        except FileNotFoundError:
            path_stats = None

        # Check to make sure it still looks locked before we unlink.
        if path_stats is None:
            logger.error(
                "PID %d had mutex %s disappear while locked! Mutex system is not working!",
                os.getpid(),
                lock_filename,
            )
        elif (
            fd_stats.st_dev != path_stats.st_dev or fd_stats.st_ino != path_stats.st_ino
        ):
            logger.error(
                "PID %d had mutex %s get replaced while locked! Mutex system is not working!",
                os.getpid(),
                lock_filename,
            )

        if path_stats is not None:
            try:
                # Unlink the file
                os.unlink(lock_filename)
            except FileNotFoundError:
                logger.error(
                    "PID %d had mutex %s disappear between stat and unlink while unlocking! Mutex system is not working!",
                    os.getpid(),
                    lock_filename,
                )

        # Note that we are unlinking it and then unlocking it; a lot of people
        # might have opened it before we unlinked it and will wake up when they
        # get the worthless lock on the now-unlinked file. We have to do some
        # stat gymnastics above to work around this.
        safe_unlock_and_close(fd)


class LastProcessStandingArena:
    """
    Class that lets a bunch of processes detect and elect a last process
    standing.

    Process enter and leave (sometimes due to sudden existence failure). We
    guarantee that the last process to leave, if it leaves properly, will get a
    chance to do some cleanup. If new processes try to enter during the
    cleanup, they will be delayed until after the cleanup has happened and the
    previous "last" process has finished leaving.

    The user is responsible for making sure you always leave if you enter!
    Consider using a try/finally; this class is not a context manager.
    """

    def __init__(self, base_dir: StrPath, name: str) -> None:
        """
        Connect to the arena specified by the given base_dir and name.

        Any process that can access base_dir, in any container, can connect to
        the arena. Many arenas can be active with different names.

        Doesn't enter or leave the arena.

        :param base_dir: Base directory to work in. Defines the shared namespace.
        :param name: Name of the arena. Must be a permissible path component.
        """

        # Save the base_dir which namespaces everything
        self.base_dir = str(base_dir)

        # We need a mutex name to allow only one process to be entering or
        # leaving at a time.
        self.mutex = name + "-arena-lock"

        # We need a way to track who is actually in, and who was in but died.
        # So everybody gets a locked file (again).
        # TODO: deduplicate with the similar logic for process names, and also
        # deferred functions.
        self.lockfileDir = os.path.join(base_dir, name + "-arena-members")

        # When we enter the arena, we fill this in with the FD of the locked
        # file that represents our presence.
        self.lockfileFD = None
        # And we fill this in with the file name
        self.lockfileName = None

    def enter(self) -> None:
        """
        This process is entering the arena. If cleanup is in progress, blocks
        until it is finished.

        You may not enter the arena again before leaving it.
        """

        logger.debug("Joining arena %s", self.lockfileDir)

        # Make sure we're not in it already.
        if self.lockfileName is not None or self.lockfileFD is not None:
            raise RuntimeError("A process is already in the arena")

        with global_mutex(self.base_dir, self.mutex):
            # Now nobody else should also be trying to join or leave.

            try:
                # Make sure the lockfile directory exists.
                os.mkdir(self.lockfileDir)
            except FileExistsError:
                pass
            except Exception as e:
                raise RuntimeError(
                    "Could not make lock file directory " + self.lockfileDir
                ) from e

            # Make ourselves a file in it and lock it to prove we are alive.
            try:
                self.lockfileFD, self.lockfileName = tempfile.mkstemp(dir=self.lockfileDir)  # type: ignore
            except Exception as e:
                raise RuntimeError(
                    "Could not make lock file in " + self.lockfileDir
                ) from e
            # Nobody can see it yet, so lock it right away
            safe_lock(self.lockfileFD)  # type: ignore

            # Now we're properly in, so release the global mutex

        logger.debug("Now in arena %s", self.lockfileDir)

    def leave(self) -> Iterator[bool]:
        """
        This process is leaving the arena. If this process happens to be the
        last process standing, yields something, with other processes blocked
        from joining the arena until the loop body completes and the process
        has finished leaving. Otherwise, does not yield anything.

        Should be used in a loop:

            for _ in arena.leave():
                # If we get here, we were the last process. Do the cleanup
                pass
        """

        # Make sure we're in it to start.
        if self.lockfileName is None or self.lockfileFD is None:
            raise RuntimeError("This process is not in the arena.")

        logger.debug("Leaving arena %s", self.lockfileDir)

        with global_mutex(self.base_dir, self.mutex):
            # Now nobody else should also be trying to join or leave.

            # Take ourselves out.
            try:
                os.unlink(self.lockfileName)
            except:
                pass
            self.lockfileName = None
            safe_unlock_and_close(self.lockfileFD)
            self.lockfileFD = None

            for item in os.listdir(self.lockfileDir):
                # There is someone claiming to be here. Are they alive?
                full_path = os.path.join(self.lockfileDir, item)

                try:
                    fd = os.open(full_path, os.O_RDONLY)
                except OSError as e:
                    # suddenly file doesnt exist on network file system?
                    continue

                try:
                    safe_lock(fd, block=False, shared=True)
                except OSError as e:
                    if e.errno in (errno.EACCES, errno.EAGAIN):
                        # Could not lock. It's alive!
                        break
                    else:
                        # Something else is wrong
                        os.close(fd)
                        raise
                else:
                    # Could lock. Process is dead.
                    try:
                        os.remove(full_path)
                    except:
                        pass
                    safe_unlock_and_close(fd)
                    # Continue with the loop normally.
            else:
                # Nothing alive was found. Nobody will come in while we hold
                # the global mutex, so we are the Last Process Standing.
                logger.debug(
                    "We are the Last Process Standing in arena %s", self.lockfileDir
                )
                yield True

                try:
                    # Delete the arena directory so as to leave nothing behind.
                    os.rmdir(self.lockfileDir)
                except:
                    logger.warning(
                        "Could not clean up arena %s completely: %s",
                        self.lockfileDir,
                        traceback.format_exc(),
                    )

            # Now we're done, whether we were the last one or not, and can
            # release the mutex.

        logger.debug("Now out of arena %s", self.lockfileDir)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import logging
import os
import socket
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Optional, cast
from urllib.parse import ParseResult

# To import toil.lib.aws.session, the AWS libraries must be installed 
from toil.lib.aws import AWSRegionName, AWSServerErrors, session
from toil.lib.conversions import strtobool
from toil.lib.memoize import memoize
from toil.lib.misc import printq
from toil.lib.retry import (
    DEFAULT_DELAYS,
    DEFAULT_TIMEOUT,
    get_error_code,
    get_error_status,
    old_retry,
    retry,
)

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3ServiceResource
    from mypy_boto3_s3.service_resource import Bucket
    from mypy_boto3_s3.service_resource import Object as S3Object
    from mypy_boto3_sdb.type_defs import AttributeTypeDef

from botocore.exceptions import ClientError, EndpointConnectionError

logger = logging.getLogger(__name__)

# These are error codes we expect from AWS if we are making requests too fast.
# https://github.com/boto/botocore/blob/49f87350d54f55b687969ec8bf204df785975077/botocore/retries/standard.py#L316
THROTTLED_ERROR_CODES = [
    "Throttling",
    "ThrottlingException",
    "ThrottledException",
    "RequestThrottledException",
    "TooManyRequestsException",
    "ProvisionedThroughputExceededException",
    "TransactionInProgressException",
    "RequestLimitExceeded",
    "BandwidthLimitExceeded",
    "LimitExceededException",
    "RequestThrottled",
    "SlowDown",
    "PriorRequestNotComplete",
    "EC2ThrottledException",
]


@retry(errors=[AWSServerErrors])
def delete_sdb_domain(
    sdb_domain_name: str, region: Optional[str] = None, quiet: bool = True
) -> None:
    sdb_client = session.client("sdb", region_name=region)
    sdb_client.delete_domain(DomainName=sdb_domain_name)
    printq(f'SBD Domain: "{sdb_domain_name}" successfully deleted.', quiet)


def connection_reset(e: Exception) -> bool:
    """
    Return true if an error is a connection reset error.
    """
    # For some reason we get 'error: [Errno 104] Connection reset by peer' where the
    # English description suggests that errno is 54 (ECONNRESET) while the actual
    # errno is listed as 104. To be safe, we check for both:
    return isinstance(e, socket.error) and e.errno in (errno.ECONNRESET, 104)


def connection_error(e: Exception) -> bool:
    """
    Return True if an error represents a failure to make a network connection.
    """
    return connection_reset(e) or isinstance(e, EndpointConnectionError)


# TODO: Replace with: @retry and ErrorCondition
def retryable_s3_errors(e: Exception) -> bool:
    """
    Return true if this is an error from S3 that looks like we ought to retry our request.
    """
    return (
        connection_error(e)
        or (isinstance(e, ClientError) and get_error_status(e) in (429, 500))
        or (isinstance(e, ClientError) and get_error_code(e) in THROTTLED_ERROR_CODES)
        # boto3 errors
        or (isinstance(e, ClientError) and get_error_code(e) in THROTTLED_ERROR_CODES)
        or (isinstance(e, ClientError) and "BucketNotEmpty" in str(e))
        or (
            isinstance(e, ClientError)
            and e.response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 409
            and "try again" in str(e)
        )
        or (
            isinstance(e, ClientError)
            and e.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
            in (404, 429, 500, 502, 503, 504)
        )
    )


def retry_s3(
    delays: Iterable[float] = DEFAULT_DELAYS,
    timeout: float = DEFAULT_TIMEOUT,
    predicate: Callable[[Exception], bool] = retryable_s3_errors,
) -> Iterator[ContextManager[None]]:
    """
    Retry iterator of context managers specifically for S3 operations.
    """
    return old_retry(delays=delays, timeout=timeout, predicate=predicate)


@retry(errors=[AWSServerErrors])
def delete_s3_bucket(
    s3_resource: "S3ServiceResource", bucket: str, quiet: bool = True
) -> None:
    """
    Delete the given S3 bucket.
    """
    printq(f"Deleting s3 bucket: {bucket}", quiet)

    paginator = s3_resource.meta.client.get_paginator("list_object_versions")
    try:
        for response in paginator.paginate(Bucket=bucket):
            # Versions and delete markers can both go in here to be deleted.
            # They both have Key and VersionId, but there's no shared base type
            # defined for them in the stubs to express that. See
            # <https://github.com/vemel/mypy_boto3_builder/issues/123>. So we
            # have to do gymnastics to get them into the same list.
            to_delete: list[dict[str, Any]] = cast(
                list[dict[str, Any]], response.get("Versions", [])
            ) + cast(list[dict[str, Any]], response.get("DeleteMarkers", []))
            for entry in to_delete:
                printq(
                    f"    Deleting {entry['Key']} version {entry['VersionId']}", quiet
                )
                s3_resource.meta.client.delete_object(
                    Bucket=bucket, Key=entry["Key"], VersionId=entry["VersionId"]
                )
        s3_resource.Bucket(bucket).delete()
        printq(f"\n * Deleted s3 bucket successfully: {bucket}\n\n", quiet)
    except s3_resource.meta.client.exceptions.NoSuchBucket:
        printq(f"\n * S3 bucket no longer exists: {bucket}\n\n", quiet)


def create_s3_bucket(
    s3_resource: "S3ServiceResource",
    bucket_name: str,
    region: AWSRegionName,
) -> "Bucket":
    """
    Create an AWS S3 bucket, using the given Boto3 S3 session, with the
    given name, in the given region.

    Supports the us-east-1 region, where bucket creation is special.

    *ALL* S3 bucket creation should use this function.
    """
    logger.info("Creating bucket '%s' in region %s.", bucket_name, region)
    if region == "us-east-1":  # see https://github.com/boto/boto3/issues/125
        bucket = s3_resource.create_bucket(Bucket=bucket_name)
    else:
        bucket = s3_resource.create_bucket(
            Bucket=bucket_name,
            CreateBucketConfiguration={"LocationConstraint": region},
        )
    return bucket


@retry(errors=[ClientError])
def enable_public_objects(bucket_name: str) -> None:
    """
    Enable a bucket to contain objects which are public.

    This adjusts the bucket's Public Access Block setting to not block all
    public access, and also adjusts the bucket's Object Ownership setting to a
    setting which enables object ACLs.

    Does *not* touch the *account*'s Public Access Block setting, which can
    also interfere here. That is probably best left to the account
    administrator.

    This configuration used to be the default, and is what most of Toil's code
    is written to expect, but it was changed so that new buckets default to the
    more restrictive setting
    <https://aws.amazon.com/about-aws/whats-new/2022/12/amazon-s3-automatically-enable-block-public-access-disable-access-control-lists-buckets-april-2023/>,
    with the expectation that people would write IAM policies for the buckets
    to allow public access if needed. Toil expects to be able to make arbitrary
    objects in arbitrary places public, and naming them all in an IAM policy
    would be a very awkward way to do it. So we restore the old behavior.
    """

    s3_client = session.client("s3")

    # Even though the new default is for public access to be prohibited, this
    # is implemented by adding new things attached to the bucket. If we remove
    # those things the bucket will default to the old defaults. See
    # <https://aws.amazon.com/blogs/aws/heads-up-amazon-s3-security-changes-are-coming-in-april-of-2023/>.

    # Stop blocking public access
    s3_client.delete_public_access_block(Bucket=bucket_name)

    # Stop using an ownership controls setting that prohibits ACLs.
    s3_client.delete_bucket_ownership_controls(Bucket=bucket_name)


class NoBucketLocationError(Exception):
    """
    Error to represent that we could not get a location for a bucket.
    """


def get_bucket_region(
    bucket_name: str,
    endpoint_url: Optional[str] = None,
    only_strategies: Optional[set[int]] = None,
    anonymous: Optional[bool] = None
) -> str:
    """
    Get the AWS region name associated with the given S3 bucket, or raise NoBucketLocationError.

    Does not log at info level or above when this does not work; failures are expected in some contexts.

    Takes an optional S3 API URL override.

    :param only_strategies: For testing, use only strategies with 1-based numbers in this set.

    :raises NoBucketLocationError: if the bucket's region cannot be determined
        (possibly due to lack of permissions).
    """

    config = session.ANONYMOUS_CONFIG if anonymous else None
    s3_client = session.client("s3", endpoint_url=endpoint_url, config=config)

    def attempt_get_bucket_location() -> Optional[str]:
        """
        Try and get the bucket location from the normal API call.
        """
        return s3_client.get_bucket_location(Bucket=bucket_name).get(
            "LocationConstraint", None
        )

    def attempt_get_bucket_location_from_us_east_1() -> Optional[str]:
        """
        Try and get the bucket location from the normal API call, but against us-east-1
        """
        # Sometimes we aren't allowed to GetBucketLocation. At least some of
        # the time, that's only true when we talk to whatever S3 API servers we
        # usually use, and we can get around this lack of permission by talking
        # to us-east-1 instead. We've been told that this is because us-east-1
        # is special and will answer the question when other regions won't.
        # See:
        # <https://ucsc-gi.slack.com/archives/C027D41M6UA/p1652819831740169?thread_ts=1652817377.594539&cid=C027D41M6UA>
        # It could also be because AWS open data buckets (which we tend to
        # encounter this problem for) tend to actually themselves be in
        # us-east-1.
        backup_s3_client = session.client("s3", region_name="us-east-1", config=config)
        return backup_s3_client.get_bucket_location(Bucket=bucket_name).get(
            "LocationConstraint", None
        )

    def attempt_head_bucket() -> Optional[str]:
        """
        Try and get the bucket location from calling HeadBucket and inspecting
        the headers.
        """
        # If that also doesn't work, we can try HEAD-ing the bucket and looking
        # for an 'x-amz-bucket-region' header on the response, which can tell
        # us where the bucket is. See
        # <https://github.com/aws/aws-sdk-cpp/issues/844#issuecomment-383747871>
        info = s3_client.head_bucket(Bucket=bucket_name)
        return info["ResponseMetadata"]["HTTPHeaders"]["x-amz-bucket-region"]

    # Compose a list of strategies we want to try in order, which may work.
    # None is an acceptable return type that actually means something.
    strategies: list[Callable[[], Optional[str]]] = []
    strategies.append(attempt_get_bucket_location)
    if not endpoint_url:
        # We should only try to talk to us-east-1 if we don't have a custom
        # URL.
        strategies.append(attempt_get_bucket_location_from_us_east_1)
    strategies.append(attempt_head_bucket)

    error_logs: list[tuple[int, str]] = []
    for attempt in retry_s3():
        with attempt:
            for i, strategy in enumerate(strategies):
                if only_strategies is not None and i + 1 not in only_strategies:
                    # We want to test running without this strategy.
                    continue
                try:
                    location = bucket_location_to_region(strategy())
                    logger.debug("Got bucket location from strategy %d", i + 1)
                    return location
                except ClientError as e:
                    if get_error_code(e) == "AccessDenied" and not endpoint_url:
                        logger.debug(
                            "Strategy %d to get bucket location did not work: %s",
                            i + 1,
                            e,
                        )
                        error_logs.append((i + 1, str(e)))
                        last_error: Exception = e
                        # We were blocked with this strategy. Move on to the
                        # next strategy which might work.
                        continue
                    else:
                        raise
                except KeyError as e:
                    # If we get a weird head response we will have a KeyError
                    logger.debug(
                        "Strategy %d to get bucket location did not work: %s", i + 1, e
                    )
                    error_logs.append((i + 1, str(e)))
                    last_error = e

    error_messages = []
    for rank, message in error_logs:
        error_messages.append(
            f"Strategy {rank} failed to get bucket location because: {message}"
        )
    # If we get here we ran out of attempts.
    raise NoBucketLocationError(
        "Could not get bucket location: " + "\n".join(error_messages)
    ) from last_error

@memoize
def get_bucket_region_if_available(
    bucket_name: str,
    endpoint_url: Optional[str] = None,
    only_strategies: Optional[set[int]] = None,
    anonymous: Optional[bool] = None
) -> Optional[str]:
    """
    Get the AWS region name associated with the given S3 bucket, or return None.

    Caches results, so may not return the location for a bucket that has been
    created but was previously observed to be nonexistent.

    :param only_strategies: For testing, use only strategies with 1-based numbers in this set.
    """

    try:
        return get_bucket_region(bucket_name, endpoint_url, only_strategies, anonymous)
    except Exception as e:
        if isinstance(e, NoBucketLocationError) or (isinstance(e, ClientError) and get_error_status(e) == 403):
            # We can't know
            return None
        else:
            raise

def region_to_bucket_location(region: str) -> str:
    return "" if region == "us-east-1" else region


def bucket_location_to_region(location: Optional[str]) -> str:
    return "us-east-1" if location == "" or location is None else location


def get_object_for_url(url: ParseResult, existing: Optional[bool] = None, anonymous: Optional[bool] = None) -> "S3Object":
    """
    Extracts a key (object) from a given parsed s3:// URL.

    If existing is true and the object does not exist, raises FileNotFoundError.

    :param bool existing: If True, key is expected to exist. If False, key is expected not to
            exists and it will be created. If None, the key will be created if it doesn't exist.

    :raises FileNotFoundError: when existing is True and the object does not exist.
    :raises RuntimeError: when existing is False but the object exists.
    :raises PermissionError: when we are not authorized to look at the object.
    """

    key_name = url.path[1:]
    bucket_name = url.netloc

    # Decide if we need to override Boto's built-in URL here.
    endpoint_url: Optional[str] = None
    host = os.environ.get("TOIL_S3_HOST", None)
    port = os.environ.get("TOIL_S3_PORT", None)
    protocol = "https"
    if strtobool(os.environ.get("TOIL_S3_USE_SSL", 'True')) is False:
        protocol = "http"
    if host:
        endpoint_url = f"{protocol}://{host}" + f":{port}" if port else ""

    # TODO: OrdinaryCallingFormat equivalent in boto3?
    # if botoargs:
    #     botoargs['calling_format'] = boto.s3.connection.OrdinaryCallingFormat()
    
    config = session.ANONYMOUS_CONFIG if anonymous else None
    # Get the bucket's region to avoid a redirect per request.
    # Cache the result
    region = get_bucket_region_if_available(bucket_name, endpoint_url=endpoint_url, anonymous=anonymous)
    if region is not None:
        s3 = session.resource("s3", region_name=region, endpoint_url=endpoint_url, config=config)
    else:
        # We can't get the bucket location, perhaps because we don't have
        # permission to do that.
        logger.debug("Couldn't get bucket location")
        logger.debug("Fall back to not specifying location")
        s3 = session.resource("s3", endpoint_url=endpoint_url, config=config)

    obj = s3.Object(bucket_name, key_name)
    objExists = True

    try:
        obj.load()
    except ClientError as e:
        if get_error_status(e) == 404:
            objExists = False
        elif get_error_status(e) == 403:
            raise PermissionError(
                f"Key '{key_name}' is not accessible in bucket '{bucket_name}'."
            ) from e
        else:
            raise
    if existing is True and not objExists:
        raise FileNotFoundError(
            f"Key '{key_name}' does not exist in bucket '{bucket_name}'."
        )
    elif existing is False and objExists:
        raise RuntimeError(f"Key '{key_name}' exists in bucket '{bucket_name}'.")

    if not objExists:
        try:
            obj.put()  # write an empty file
        except ClientError as e:
            if get_error_status(e) == 403:
                raise PermissionError(
                    f"Key '{key_name}' is not writable in bucket '{bucket_name}'."
                ) from e
            else:
                raise
    return obj


@retry(errors=[AWSServerErrors])
def list_objects_for_url(url: ParseResult, anonymous: Optional[bool] = None) -> list[str]:
    """
    Extracts a key (object) from a given parsed s3:// URL. The URL will be
    supplemented with a trailing slash if it is missing.

    :raises PermissionError: when we are not authorized to do the list operation.
    """

    key_name = url.path[1:]
    bucket_name = url.netloc

    if key_name != "" and not key_name.endswith("/"):
        # Make sure to put the trailing slash on the key, or else we'll see
        # a prefix of just it.
        key_name = key_name + "/"

    # Decide if we need to override Boto's built-in URL here.
    # TODO: Deduplicate with get_object_for_url, or push down into session module
    endpoint_url: Optional[str] = None
    host = os.environ.get("TOIL_S3_HOST", None)
    port = os.environ.get("TOIL_S3_PORT", None)
    protocol = "https"
    if strtobool(os.environ.get("TOIL_S3_USE_SSL", 'True')) is False:
        protocol = "http"
    if host:
        endpoint_url = f"{protocol}://{host}" + f":{port}" if port else ""
    
    config = session.ANONYMOUS_CONFIG if anonymous else None
    client = session.client("s3", endpoint_url=endpoint_url, config=config)

    listing = []
    
    try:
        paginator = client.get_paginator("list_objects_v2")
        result = paginator.paginate(Bucket=bucket_name, Prefix=key_name, Delimiter="/")
        for page in result:
            if "CommonPrefixes" in page:
                for prefix_item in page["CommonPrefixes"]:
                    listing.append(prefix_item["Prefix"][len(key_name) :])
            if "Contents" in page:
                for content_item in page["Contents"]:
                    if content_item["Key"] == key_name:
                        # Ignore folder name itself
                        continue
                    listing.append(content_item["Key"][len(key_name) :])
    except ClientError as e:
        if get_error_status(e) == 403:
            raise PermissionError(
                f"Prefix '{key_name}' is not authorized to be listed in bucket '{bucket_name}'."
            ) from e
        else:
            raise


    logger.debug("Found in %s items: %s", url, listing)
    return listing


def flatten_tags(tags: dict[str, str]) -> list[dict[str, str]]:
    """
    Convert tags from a key to value dict into a list of 'Key': xxx, 'Value': xxx dicts.
    """
    return [{"Key": k, "Value": v} for k, v in tags.items()]


def boto3_pager(
    requestor_callable: Callable[..., Any], result_attribute_name: str, **kwargs: Any
) -> Iterable[Any]:
    """
    Yield all the results from calling the given Boto 3 method with the
    given keyword arguments, paging through the results using the Marker or
    NextToken, and fetching out and looping over the list in the response
    with the given attribute name.
    """

    # Recover the Boto3 client, and the name of the operation
    client = requestor_callable.__self__  # type: ignore[attr-defined]
    op_name = requestor_callable.__name__

    # grab a Boto 3 built-in paginator. See
    # <https://boto3.amazonaws.com/v1/documentation/api/latest/guide/paginators.html>
    paginator = client.get_paginator(op_name)

    for page in paginator.paginate(**kwargs):
        # Invoke it and go through the pages, yielding from them
        yield from page.get(result_attribute_name, [])


def get_item_from_attributes(attributes: list["AttributeTypeDef"], name: str) -> Any:
    """
    Given a list of attributes, find the attribute associated with the name and return its corresponding value.

    The `attribute_list` will be a list of TypedDict's (which boto3 SDB functions commonly return),
    where each TypedDict has a "Name" and "Value" key value pair.
    This function grabs the value out of the associated TypedDict.

    If the attribute with the name does not exist, the function will return None.

    :param attributes: list of attributes
    :param name: name of the attribute
    :return: value of the attribute
    """
    return next(
        (attribute["Value"] for attribute in attributes if attribute["Name"] == name),
        None,
    )



from toil.common import Toil
from toil.job import Job


class HelloWorld(Job):
    def __init__(self, message):
        Job.__init__(self)
        self.message = message

    def run(self, fileStore):
        return "Hello, world!, here's a message: %s" % self.message


if __name__ == "__main__":
    parser = Job.Runner.getDefaultArgumentParser()
    options = parser.parse_args()
    options.logLevel = "OFF"
    options.clean = "always"

    hello_job = HelloWorld("Woot")

    with Toil(options) as toil:
        print(toil.start(hello_job))



# Copyright (C) 2015-2022 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from pathlib import Path
from threading import Thread, current_thread
from typing import NoReturn

from toil.batchSystems.abstractBatchSystem import BatchJobExitReason
from toil.bus import (
    JobCompletedMessage,
    JobIssuedMessage,
    MessageBus,
    replay_message_bus,
)
from toil.common import Toil
from toil.exceptions import FailedJobsException
from toil.job import Job

logger = logging.getLogger(__name__)


class TestMessageBus:

    def test_enum_ints_in_file(self, tmp_path: Path) -> None:
        """
        Make sure writing bus messages to files works with enums.
        """
        bus_file = tmp_path / "bus"

        bus = MessageBus()
        # Connect the handler and hold the result to protect it from GC
        handler_to_keep_alive = bus.connect_output_file(bus_file)
        logger.debug("Connected bus to file %s", bus_file)
        bus.publish(JobCompletedMessage("blah", "blah", BatchJobExitReason.MEMLIMIT))
        # Make sure stuff goes away in the right order
        del handler_to_keep_alive
        del bus

        for line in open(bus_file):
            logger.debug("Bus line: %s", line)

        replay = replay_message_bus(bus_file)
        assert len(replay) > 0

    def test_cross_thread_messaging(self) -> None:
        """
        Make sure message bus works across threads.
        """

        main_thread = current_thread()

        bus = MessageBus()

        message_count = 0

        # We will count the messages going to the handler
        def handler(received: JobIssuedMessage) -> None:
            """
            Handle a message in the main thread.
            """
            # Message should always arrive in the main thread.
            nonlocal message_count
            logger.debug("Got message: %s", received)
            assert current_thread() == main_thread
            message_count += 1

        bus.subscribe(JobIssuedMessage, handler)

        # And we will collect them in an inbox
        box = bus.connect([JobIssuedMessage])

        bus.publish(JobIssuedMessage("FromMainThread", "", 0))

        def send_thread_message() -> None:
            """
            Publish a message from some other thread.
            """
            bus.publish(JobIssuedMessage("FromOtherThread", "", 1))

        other_threads = []
        for _ in range(10):
            # Make a bunch of threads to send messages
            other_threads.append(Thread(target=send_thread_message))
        for t in other_threads:
            # Start all the threads
            t.start()
        for t in other_threads:
            # Wait for them to finish
            t.join()

        # We should ge tone message per thread, plus our own
        assert box.count(JobIssuedMessage) == 11
        # And having polled for those, our handler should have run
        assert message_count == 11

    def test_restart_without_bus_path(self, tmp_path: Path) -> None:
        """
        Test the ability to restart a workflow when the message bus path used
        by the previous attempt is gone.
        """
        temp_dir = tmp_path / "tempDir"
        temp_dir.mkdir()
        job_store = tmp_path / "jobstore"

        bus_holder_dir = temp_dir / "bus_holder"
        bus_holder_dir.mkdir()

        start_options = Job.Runner.getDefaultOptions(job_store)
        start_options.logLevel = "DEBUG"
        start_options.retryCount = 0
        start_options.clean = "never"
        start_options.write_messages = str(bus_holder_dir / "messagebus.txt")

        root = Job.wrapJobFn(failing_job_fn)

        try:
            with Toil(start_options) as toil:
                # Run once and observe a failed job
                toil.start(root)
        except FailedJobsException:
            pass

        logger.info("First attempt successfully failed, removing message bus log")

        # Get rid of the bus
        os.unlink(start_options.write_messages)
        bus_holder_dir.rmdir()

        logger.info("Making second attempt")

        # Set up options without a specific bus path
        restart_options = Job.Runner.getDefaultOptions(job_store)
        restart_options.logLevel = "DEBUG"
        restart_options.retryCount = 0
        restart_options.clean = "never"
        restart_options.restart = True

        try:
            with Toil(restart_options) as toil:
                # Run again and observe a failed job (and not a failure to start)
                toil.restart()
        except FailedJobsException:
            pass

        logger.info("Second attempt successfully failed")


def failing_job_fn(job: Job) -> NoReturn:
    """
    This function is guaranteed to fail.
    """
    raise RuntimeError("Job attempted to run but failed")



# Copyright (C) 2015-2025 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Generator
from pathlib import Path
import logging
import time

from toil.lib.history import HistoryManager

import pytest

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class TestHistory:
    """
    Tests for Toil history tracking.

    Each test gets its own history database.
    """

    @pytest.fixture(autouse=True, scope="function")
    def private_history_manager(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> Generator[None]:
        try:
            with monkeypatch.context() as m:
                m.setattr(
                    HistoryManager,
                    "database_path_override",
                    str(tmp_path / "test-db.sqlite"),
                )
                m.setattr(HistoryManager, "enabled", lambda: True)
                m.setattr(HistoryManager, "enabled_job", lambda: True)
                yield
        finally:
            pass  # no cleanup needed

    def make_fake_workflow(self, workflow_id: str) -> None:
        # Make a fake workflow
        workflow_jobstore_spec = "file:/tmp/tree"
        HistoryManager.record_workflow_creation(workflow_id, workflow_jobstore_spec)
        workflow_name = "SuperCoolWF"
        workflow_trs_spec = "#wf:v1"
        HistoryManager.record_workflow_metadata(
            workflow_id, workflow_name, workflow_trs_spec
        )

        # Give it a job
        workflow_attempt_number = 1
        job_name = "DoThing"
        succeeded = True
        start_time = time.time()
        runtime = 0.1
        HistoryManager.record_job_attempt(
            workflow_id,
            workflow_attempt_number,
            job_name,
            succeeded,
            start_time,
            runtime,
        )

        # Give it a workflow attempt with the same details.
        HistoryManager.record_workflow_attempt(
            workflow_id,
            workflow_attempt_number,
            succeeded,
            start_time,
            runtime,
        )

    def test_history_submittable_detection(self) -> None:
        """
        Make sure that a submittable workflow shows up as such before
        submission and doesn't afterward.
        """
        workflow_id = "123"
        self.make_fake_workflow(workflow_id)
        workflow_attempt_number = 1

        # Make sure we have data
        assert HistoryManager.count_workflows() == 1
        assert HistoryManager.count_workflow_attempts() == 1
        assert HistoryManager.count_job_attempts() == 1

        # Make sure we see it as submittable
        submittable_workflow_attempts = (
            HistoryManager.get_submittable_workflow_attempts()
        )
        assert len(submittable_workflow_attempts) == 1

        # Make sure we see its jobs as submittable
        with_submittable_job_attempts = (
            HistoryManager.get_workflow_attempts_with_submittable_job_attempts()
        )
        assert len(with_submittable_job_attempts) == 1

        # Make sure we actually see the job
        submittable_job_attempts = HistoryManager.get_unsubmitted_job_attempts(
            workflow_id, workflow_attempt_number
        )
        assert len(submittable_job_attempts) == 1

        # Pretend we submitted them.
        HistoryManager.mark_job_attempts_submitted(
            [j.id for j in submittable_job_attempts]
        )
        HistoryManager.mark_workflow_attempt_submitted(
            workflow_id, workflow_attempt_number
        )

        # Make sure they are no longer matching
        assert len(HistoryManager.get_submittable_workflow_attempts()) == 0
        assert (
            len(HistoryManager.get_workflow_attempts_with_submittable_job_attempts())
            == 0
        )
        assert (
            len(
                HistoryManager.get_unsubmitted_job_attempts(
                    workflow_id, workflow_attempt_number
                )
            )
            == 0
        )

        # Make sure we still have data
        assert HistoryManager.count_workflows() == 1
        assert HistoryManager.count_workflow_attempts() == 1
        assert HistoryManager.count_job_attempts() == 1

    def test_history_deletion(self) -> None:
        workflow_id = "123"
        self.make_fake_workflow(workflow_id)
        workflow_attempt_number = 1

        # Make sure we can see the workflow for deletion by age but not by done-ness
        assert len(HistoryManager.get_oldest_workflow_ids()) == 1
        assert len(HistoryManager.get_fully_submitted_workflow_ids()) == 0

        # Pretend we submitted the workflow.
        HistoryManager.mark_job_attempts_submitted(
            [
                j.id
                for j in HistoryManager.get_unsubmitted_job_attempts(
                    workflow_id, workflow_attempt_number
                )
            ]
        )
        HistoryManager.mark_workflow_attempt_submitted(
            workflow_id, workflow_attempt_number
        )

        # Make sure we can see the workflow for deletion by done-ness
        assert len(HistoryManager.get_fully_submitted_workflow_ids()) == 1

        # Add a new workflow
        other_workflow_id = "456"
        self.make_fake_workflow(other_workflow_id)

        # Make sure we can see the both for deletion by age but only one by done-ness
        assert len(HistoryManager.get_oldest_workflow_ids()) == 2
        assert len(HistoryManager.get_fully_submitted_workflow_ids()) == 1

        # Make sure the older workflow is first.
        assert HistoryManager.get_oldest_workflow_ids() == [
            workflow_id,
            other_workflow_id,
        ]

        # Delete the new workflow
        HistoryManager.delete_workflow(other_workflow_id)

        # Make sure we can see the old one
        assert HistoryManager.get_oldest_workflow_ids() == [workflow_id]
        assert HistoryManager.get_fully_submitted_workflow_ids() == [workflow_id]

        # Delete the old workflow
        HistoryManager.delete_workflow(workflow_id)

        # Make sure we have no data
        assert HistoryManager.count_workflows() == 0
        assert HistoryManager.count_workflow_attempts() == 0
        assert HistoryManager.count_job_attempts() == 0

    def test_history_size_limit(self) -> None:
        """
        Make sure the database size can be controlled.
        """

        for workflow_id in (
            "WorkflowThatTakesUpSomeSpace,ActuallyMoreThanTheLaterOnesTake" + str(i)
            for i in range(10)
        ):
            self.make_fake_workflow(workflow_id)

        # We should see the workflows.
        assert HistoryManager.count_workflows() == 10
        # And they take up space.
        small_size = HistoryManager.get_database_byte_size()
        assert small_size > 0

        # Add a bunch more
        for workflow_id in ("WorkflowThatTakesUpSpace" + str(i) for i in range(50)):
            self.make_fake_workflow(workflow_id)

        # We should see that this is now a much larger database
        large_size = HistoryManager.get_database_byte_size()
        logger.info("Increased database size from %s to %s", small_size, large_size)
        large_size > small_size

        # We should be able to shrink it back down
        HistoryManager.enforce_byte_size_limit(small_size)

        reduced_size = HistoryManager.get_database_byte_size()
        logger.info("Decreased database size from %s to %s", large_size, reduced_size)
        # The database should be small enough
        reduced_size <= small_size
        # There should still be some workflow attempts left in the smaller database (though probably not the first ones)
        remaining_workflows = HistoryManager.count_workflows()
        logger.info("Still have %s workflows", remaining_workflows)
        assert remaining_workflows > 0



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Iterable, Generator
import argparse
import fcntl
import itertools
import logging
import os
from pathlib import Path
import subprocess
import sys
import tempfile
import textwrap
import time
from abc import ABCMeta, abstractmethod
from fractions import Fraction
from unittest import skipIf
from typing import Optional, Any, TYPE_CHECKING

from toil.batchSystems.abstractBatchSystem import (
    AbstractBatchSystem,
    BatchSystemSupport,
    InsufficientSystemResources,
)

# Don't import any batch systems here that depend on extras
# in order to import properly. Import them later, in tests
# protected by annotations.
from toil.batchSystems.mesos.test import MesosTestSupport
from toil.batchSystems.registry import (
    add_batch_system_factory,
    get_batch_system,
    get_batch_systems,
    restore_batch_system_plugin_state,
    save_batch_system_plugin_state,
)
from toil.batchSystems.singleMachine import SingleMachineBatchSystem
from toil.common import Config, Toil
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.job import Job, JobDescription, Requirer, ServiceHostJob
from toil.lib.misc import StrPath
from toil.lib.retry import retry_flaky_test
from toil.lib.threading import cpu_count
from toil.test import (
    ToilTest,
    needs_aws_batch,
    needs_aws_s3,
    needs_fetchable_appliance,
    needs_gridengine,
    needs_htcondor,
    needs_kubernetes,
    needs_kubernetes_installed,
    needs_lsf,
    needs_mesos,
    needs_slurm,
    needs_torque,
    slow,
    pslow,
    pneeds_mesos,
)

import pytest

if TYPE_CHECKING:
    from toil.batchSystems.mesos.batchSystem import MesosBatchSystem


logger = logging.getLogger(__name__)

# How many cores should be utilized by this test. The test will fail if the running system
# doesn't have at least that many cores.

numCores = 2

preemptible = False

# Since we aren't always attaching the config to the jobs for these tests, we
# need to use fully specified requirements.
defaultRequirements = dict(
    memory=int(100e6), cores=1, disk=1000, preemptible=preemptible, accelerators=[]
)


class BatchSystemPluginTest(ToilTest):
    """
    Class for testing batch system plugin functionality.
    """

    def setUp(self) -> None:
        # Save plugin state so our plugin doesn't stick around after the test
        # (and create duplicate options)
        self.__state = save_batch_system_plugin_state()
        super().setUp()

    def tearDown(self) -> None:
        # Restore plugin state
        restore_batch_system_plugin_state(self.__state)
        super().tearDown()

    def test_add_batch_system_factory(self) -> None:
        def test_batch_system_factory() -> type[SingleMachineBatchSystem]:
            # TODO: Adding the same batch system under multiple names means we
            # can't actually create Toil options, because each version tries to
            # add its arguments.
            return SingleMachineBatchSystem

        add_batch_system_factory("testBatchSystem", test_batch_system_factory)
        assert "testBatchSystem" in get_batch_systems()
        assert get_batch_system("testBatchSystem") == SingleMachineBatchSystem


class hidden:
    """
    Hide abstract base class from unittest's test case loader

    http://stackoverflow.com/questions/1323455/python-unit-test-with-base-and-sub-class#answer-25695512
    """
    class AbstractBatchSystemTest(ToilTest, metaclass=ABCMeta):
        """
        A base test case with generic tests that every batch system should pass.

        Cannot assume that the batch system actually executes commands on the local machine/filesystem.
        """

        @abstractmethod
        def createBatchSystem(self) -> AbstractBatchSystem:
            raise NotImplementedError

        def supportsWallTime(self) -> bool:
            return False

        @classmethod
        def createConfig(cls) -> Config:
            """
            Returns a dummy config for the batch system tests.  We need a workflowID to be set up
            since we are running tests without setting up a jobstore. This is the class version
            to be used when an instance is not available.

            :rtype: toil.common.Config
            """
            config = Config()
            from uuid import uuid4

            config.workflowID = str(uuid4())
            config.cleanWorkDir = "always"
            return config

        def _createConfig(self) -> Config:
            """
            Returns a dummy config for the batch system tests.  We need a workflowID to be set up
            since we are running tests without setting up a jobstore.

            :rtype: toil.common.Config
            """
            return self.createConfig()

        def _mockJobDescription(
            self, jobStoreID: Optional[str] = None, **kwargs: Any
        ) -> JobDescription:
            """
            Create a mock-up JobDescription with the given ID and other parameters.
            """

            # TODO: Use a real unittest.Mock? For now we make a real instance and just hack it up.

            desc = JobDescription(**kwargs)
            # Normally we can't pass in an ID, and the job serialization logic
            # takes care of filling it in. We set it here.
            if jobStoreID is not None:
                desc.jobStoreID = jobStoreID

            return desc

        @classmethod
        def setUpClass(cls) -> None:
            super().setUpClass()
            logging.basicConfig(level=logging.DEBUG)

        def setUp(self) -> None:
            super().setUp()
            self.config = self._createConfig()
            self.batchSystem = self.createBatchSystem()
            self.tempDir = self._createTempDir("testFiles")

        def tearDown(self) -> None:
            self.batchSystem.shutdown()
            super().tearDown()

        def get_max_startup_seconds(self) -> int:
            """
            Get the number of seconds this test ought to wait for the first job to run.
            Some batch systems may need time to scale up.
            """
            return 120

        def test_available_cores(self) -> None:
            self.assertTrue(cpu_count() >= numCores)

        @retry_flaky_test(prepare=[tearDown, setUp])
        def test_run_jobs(self) -> None:
            jobDesc1 = self._mockJobDescription(
                jobName="test1",
                unitName=None,
                jobStoreID="1",
                requirements=defaultRequirements,
            )
            jobDesc2 = self._mockJobDescription(
                jobName="test2",
                unitName=None,
                jobStoreID="2",
                requirements=defaultRequirements,
            )
            job1 = self.batchSystem.issueBatchJob("sleep 1000", jobDesc1)
            job2 = self.batchSystem.issueBatchJob("sleep 1000", jobDesc2)

            issuedIDs = self._waitForJobsToIssue(2)
            self.assertEqual(set(issuedIDs), {job1, job2})

            # Now at some point we want these jobs to become running
            # But since we may be testing against a live cluster (Kubernetes)
            # we want to handle weird cases and high cluster load as much as we can.

            # Wait a bit for any Dockers to download and for the
            # jobs to have a chance to start.
            # TODO: We insist on neither of these ever finishing when we test
            # getUpdatedBatchJob, and the sleep time is longer than the time we
            # should spend waiting for both to start, so if our cluster can
            # only run one job at a time, we will fail the test.
            runningJobIDs = self._waitForJobsToStart(
                2, tries=self.get_max_startup_seconds()
            )
            self.assertEqual(set(runningJobIDs), {job1, job2})

            # Killing the jobs instead of allowing them to complete means this test can run very
            # quickly if the batch system issues and starts the jobs quickly.
            self.batchSystem.killBatchJobs([job1, job2])
            self.assertEqual({}, self.batchSystem.getRunningBatchJobIDs())

            # Issue a job and then allow it to finish by itself, causing it to be added to the
            # updated jobs queue.
            # We would like to have this touch something on the filesystem and
            # then check for it having happened, but we can't guarantee that
            # the batch system will run against the same filesystem we are
            # looking at.
            jobDesc3 = self._mockJobDescription(
                jobName="test3",
                unitName=None,
                jobStoreID="3",
                requirements=defaultRequirements,
            )
            job3 = self.batchSystem.issueBatchJob("mktemp -d", jobDesc3)

            jobUpdateInfo = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
            assert jobUpdateInfo is not None
            jobID, exitStatus, wallTime = (
                jobUpdateInfo.jobID,
                jobUpdateInfo.exitStatus,
                jobUpdateInfo.wallTime,
            )
            logger.info(f"Third job completed: {jobID} {exitStatus} {wallTime}")

            # Since the first two jobs were killed, the only job in the updated jobs queue should
            # be job 3. If the first two jobs were (incorrectly) added to the queue, this will
            # fail with jobID being equal to job1 or job2.
            self.assertEqual(jobID, job3)
            self.assertEqual(exitStatus, 0)
            if self.supportsWallTime():
                assert wallTime is not None
                self.assertTrue(wallTime > 0)
            else:
                self.assertIsNone(wallTime)
            # TODO: Work out a way to check if the job we asked to run actually ran.
            # Don't just believe the batch system, but don't assume it ran on this machine either.
            self.assertFalse(self.batchSystem.getUpdatedBatchJob(0))

            # Make sure killBatchJobs can handle jobs that don't exist
            self.batchSystem.killBatchJobs([10])

        def test_set_env(self) -> None:
            # Start with a relatively safe script
            script_shell = (
                'if [ "x${FOO}" == "xbar" ] ; then exit 23 ; else exit 42 ; fi'
            )

            # Escape the semicolons
            script_protected = script_shell.replace(";", r"\;")

            # Turn into a string which convinces bash to take all args and paste them back together and run them
            command = 'bash -c "\\${@}" bash eval ' + script_protected
            jobDesc4 = self._mockJobDescription(
                jobName="test4",
                unitName=None,
                jobStoreID="4",
                requirements=defaultRequirements,
            )
            job4 = self.batchSystem.issueBatchJob(command, jobDesc4)
            jobUpdateInfo = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
            assert jobUpdateInfo is not None
            jobID, exitStatus, wallTime = (
                jobUpdateInfo.jobID,
                jobUpdateInfo.exitStatus,
                jobUpdateInfo.wallTime,
            )
            self.assertEqual(exitStatus, 42)
            self.assertEqual(jobID, job4)
            # Now set the variable and ensure that it is present
            self.batchSystem.setEnv("FOO", "bar")
            jobDesc5 = self._mockJobDescription(
                jobName="test5",
                unitName=None,
                jobStoreID="5",
                requirements=defaultRequirements,
            )
            job5 = self.batchSystem.issueBatchJob(command, jobDesc5)
            jobUpdateInfo2 = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
            assert jobUpdateInfo2 is not None
            self.assertEqual(jobUpdateInfo2.exitStatus, 23)
            self.assertEqual(jobUpdateInfo2.jobID, job5)

        def test_set_job_env(self) -> None:
            """Test the mechanism for setting per-job environment variables to batch system jobs."""
            script = 'if [ "x${FOO}" == "xbar" ] ; then exit 23 ; else exit 42 ; fi'
            command = 'bash -c "\\${@}" bash eval ' + script.replace(";", r"\;")

            # Issue a job with a job environment variable
            job_desc_6 = self._mockJobDescription(
                jobName="test6",
                unitName=None,
                jobStoreID="6",
                requirements=defaultRequirements,
            )
            job6 = self.batchSystem.issueBatchJob(
                command, job_desc_6, job_environment={"FOO": "bar"}
            )
            job_update_info = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
            assert job_update_info is not None
            self.assertEqual(job_update_info.exitStatus, 23)  # this should succeed
            self.assertEqual(job_update_info.jobID, job6)
            # Now check that the environment variable doesn't exist for other jobs
            job_desc_7 = self._mockJobDescription(
                jobName="test7",
                unitName=None,
                jobStoreID="7",
                requirements=defaultRequirements,
            )
            job7 = self.batchSystem.issueBatchJob(command, job_desc_7)
            job_update_info2 = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
            assert job_update_info2 is not None
            self.assertEqual(job_update_info2.exitStatus, 42)
            self.assertEqual(job_update_info2.jobID, job7)

        def testCheckResourceRequest(self) -> None:
            if isinstance(self.batchSystem, BatchSystemSupport):
                check_resource_request = self.batchSystem.check_resource_request
                # Assuming we have <2000 cores, this should be too many cores
                self.assertRaises(
                    InsufficientSystemResources,
                    check_resource_request,
                    Requirer(dict(memory=1000, cores=2000, disk="1G", accelerators=[])),
                )
                self.assertRaises(
                    InsufficientSystemResources,
                    check_resource_request,
                    Requirer(dict(memory=5, cores=2000, disk="1G", accelerators=[])),
                )

                # This should be too much memory
                self.assertRaises(
                    InsufficientSystemResources,
                    check_resource_request,
                    Requirer(dict(memory="5000G", cores=1, disk="1G", accelerators=[])),
                )

                # This should be too much disk
                self.assertRaises(
                    InsufficientSystemResources,
                    check_resource_request,
                    Requirer(dict(memory=5, cores=1, disk="2G", accelerators=[])),
                )

                # This should be an accelerator we don't have.
                # All the batch systems need code to know they don't have these accelerators.
                self.assertRaises(
                    InsufficientSystemResources,
                    check_resource_request,
                    Requirer(
                        dict(
                            memory=5,
                            cores=1,
                            disk=100,
                            accelerators=[{"kind": "turbo-encabulator", "count": 1}],
                        )
                    ),
                )

                # These should be missing attributes
                self.assertRaises(
                    AttributeError,
                    check_resource_request,
                    Requirer(dict(memory=5, cores=1, disk=1000)),
                )
                self.assertRaises(
                    AttributeError,
                    check_resource_request,
                    Requirer(dict(cores=1, disk=1000, accelerators=[])),
                )
                self.assertRaises(
                    AttributeError,
                    check_resource_request,
                    Requirer(dict(memory=10, disk=1000, accelerators=[])),
                )

                # This should actually work
                check_resource_request(
                    Requirer(dict(memory=10, cores=1, disk=100, accelerators=[]))
                )

        def testScalableBatchSystem(self) -> None:
            # If instance of scalable batch system
            pass

        def _waitForJobsToIssue(self, numJobs: int) -> list[int]:
            issuedIDs = []
            for it in range(20):
                issuedIDs = self.batchSystem.getIssuedBatchJobIDs()
                if len(issuedIDs) == numJobs:
                    break
                time.sleep(1)
            return issuedIDs

        def _waitForJobsToStart(self, numJobs: int, tries: int = 20) -> list[int]:
            """
            Loop until the given number of distinct jobs are in the
            running state, or until the given number of tries is exhausted
            (with 1 second polling period).

            Returns the list of IDs that are running.
            """
            runningIDs = []
            # prevent an endless loop, give it a few tries
            for it in range(tries):
                running = self.batchSystem.getRunningBatchJobIDs()
                logger.info(f"Running jobs now: {running}")
                runningIDs = list(running.keys())
                if len(runningIDs) == numJobs:
                    break
                time.sleep(1)
            return runningIDs

    class AbstractGridEngineBatchSystemTest(AbstractBatchSystemTest):
        """
        An abstract class to reduce redundancy between Grid Engine, Slurm, and other similar batch
        systems
        """

        def _createConfig(self) -> Config:
            config = super()._createConfig()
            config.statePollingWait = 0.5  # Reduce polling wait so tests run faster
            # can't use _getTestJobStorePath since that method removes the directory
            config.jobStore = "file:" + self._createTempDir("jobStore")
            return config


class AbstractBatchSystemJobTest:
    """
    An abstract base class for batch system tests that use a full Toil workflow rather
    than using the batch system directly.
    """

    cpuCount = cpu_count() if cpu_count() < 4 else 4
    allocatedCores = sorted({1, 2, cpuCount})
    sleepTime = 30

    @abstractmethod
    def getBatchSystemName(self) -> str:
        """
        :rtype: (str, AbstractBatchSystem)
        """
        raise NotImplementedError

    def getOptions(self, tempDir: Path) -> argparse.Namespace:
        """
        Configures options for Toil workflow and makes job store.
        :param str tempDir: path to test directory
        :return: Toil options object
        """
        workdir = tempDir / "workdir"
        workdir.mkdir()
        options = Job.Runner.getDefaultOptions(tempDir / "jobstore")
        options.logLevel = "DEBUG"
        options.batchSystem = self.getBatchSystemName()
        options.workDir = str(workdir)
        options.maxCores = self.cpuCount
        return options

    @pslow
    @pytest.mark.slow
    def testJobConcurrency(self, tmp_path: Path) -> None:
        """
        Tests that the batch system is allocating core resources properly for concurrent tasks.
        """
        for coresPerJob in self.allocatedCores:
            tempDir = tmp_path / f"testFiles_{coresPerJob}"
            tempDir.mkdir()
            options = self.getOptions(tempDir)

            counterPath = tempDir / "counter"
            resetCounters(counterPath)
            value, maxValue = getCounters(counterPath)
            assert (value, maxValue) == (0, 0)

            root = Job()
            for _ in range(self.cpuCount):
                root.addFollowOn(
                    Job.wrapFn(
                        measureConcurrency,
                        counterPath,
                        self.sleepTime,
                        cores=coresPerJob,
                        memory="1M",
                        disk="1Mi",
                    )
                )
            with Toil(options) as toil:
                toil.start(root)
            _, maxValue = getCounters(counterPath)
            assert maxValue == (self.cpuCount // coresPerJob)

    def test_omp_threads(self, tmp_path: Path) -> None:
        """
        Test if the OMP_NUM_THREADS env var is set correctly based on jobs.cores.
        """
        test_cases = {
            # mapping of the number of cores to the OMP_NUM_THREADS value
            0.1: "1",
            1: "1",
            2: "2",
        }

        options = self.getOptions(tmp_path)

        for cores, expected_omp_threads in test_cases.items():
            if eont := os.environ.get("OMP_NUM_THREADS"):
                expected_omp_threads = eont
                logger.info(
                    f"OMP_NUM_THREADS is set.  Using OMP_NUM_THREADS={expected_omp_threads} instead."
                )
            with Toil(options) as toil:
                output = toil.start(
                    Job.wrapFn(get_omp_threads, memory="1Mi", cores=cores, disk="1Mi")
                )
            assert output == expected_omp_threads


@needs_kubernetes
@needs_aws_s3
@needs_fetchable_appliance
class KubernetesBatchSystemTest(hidden.AbstractBatchSystemTest):
    """
    Tests against the Kubernetes batch system
    """

    def supportsWallTime(self) -> bool:
        return True

    def createBatchSystem(self) -> AbstractBatchSystem:
        # We know we have Kubernetes so we can import the batch system
        from toil.batchSystems.kubernetes import KubernetesBatchSystem

        return KubernetesBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1e9, maxDisk=2001
        )


@needs_kubernetes_installed
class KubernetesBatchSystemBenchTest(ToilTest):
    """
    Kubernetes batch system unit tests that don't need to actually talk to a cluster.
    """

    def test_preemptability_constraints(self) -> None:
        """
        Make sure we generate the right preemptability constraints.
        """

        # Make sure we can print diffs of these long strings
        self.maxDiff = 10000

        from kubernetes.client import V1PodSpec

        from toil.batchSystems.kubernetes import KubernetesBatchSystem

        normal_spec = V1PodSpec(containers=[])
        constraints = KubernetesBatchSystem.Placement()
        constraints.set_preemptible(False)
        constraints.apply(normal_spec)
        self.assertEqual(
            textwrap.dedent(
                """
        {'node_affinity': {'preferred_during_scheduling_ignored_during_execution': None,
                           'required_during_scheduling_ignored_during_execution': {'node_selector_terms': [{'match_expressions': [{'key': 'eks.amazonaws.com/capacityType',
                                                                                                                                   'operator': 'NotIn',
                                                                                                                                   'values': ['SPOT']},
                                                                                                                                  {'key': 'cloud.google.com/gke-preemptible',
                                                                                                                                   'operator': 'NotIn',
                                                                                                                                   'values': ['true']}],
                                                                                                            'match_fields': None}]}},
         'pod_affinity': None,
         'pod_anti_affinity': None}
        """
            ).strip(),
            str(normal_spec.affinity),
        )
        self.assertEqual(str(normal_spec.tolerations), "None")

        spot_spec = V1PodSpec(containers=[])
        constraints = KubernetesBatchSystem.Placement()
        constraints.set_preemptible(True)
        constraints.apply(spot_spec)
        self.assertEqual(
            textwrap.dedent(
                """
        {'node_affinity': {'preferred_during_scheduling_ignored_during_execution': [{'preference': {'match_expressions': [{'key': 'eks.amazonaws.com/capacityType',
                                                                                                                           'operator': 'In',
                                                                                                                           'values': ['SPOT']}],
                                                                                                    'match_fields': None},
                                                                                     'weight': 1},
                                                                                    {'preference': {'match_expressions': [{'key': 'cloud.google.com/gke-preemptible',
                                                                                                                           'operator': 'In',
                                                                                                                           'values': ['true']}],
                                                                                                    'match_fields': None},
                                                                                     'weight': 1}],
                           'required_during_scheduling_ignored_during_execution': None},
         'pod_affinity': None,
         'pod_anti_affinity': None}
        """
            ).strip(),
            str(spot_spec.affinity),
        )
        self.assertEqual(
            textwrap.dedent(
                """
        [{'effect': None,
         'key': 'cloud.google.com/gke-preemptible',
         'operator': None,
         'toleration_seconds': None,
         'value': 'true'}]
        """
            ).strip(),
            str(spot_spec.tolerations),
        )

    def test_label_constraints(self) -> None:
        """
        Make sure we generate the right preemptability constraints.
        """

        # Make sure we can print diffs of these long strings
        self.maxDiff = 10000

        from kubernetes.client import V1PodSpec

        from toil.batchSystems.kubernetes import KubernetesBatchSystem

        spec = V1PodSpec(containers=[])
        constraints = KubernetesBatchSystem.Placement()
        constraints.required_labels = [("GottaBeSetTo", ["This"])]
        constraints.desired_labels = [("OutghtToBeSetTo", ["That"])]
        constraints.prohibited_labels = [("CannotBe", ["ABadThing"])]
        constraints.apply(spec)
        self.assertEqual(
            textwrap.dedent(
                """
        {'node_affinity': {'preferred_during_scheduling_ignored_during_execution': [{'preference': {'match_expressions': [{'key': 'OutghtToBeSetTo',
                                                                                                                           'operator': 'In',
                                                                                                                           'values': ['That']}],
                                                                                                    'match_fields': None},
                                                                                     'weight': 1}],
                           'required_during_scheduling_ignored_during_execution': {'node_selector_terms': [{'match_expressions': [{'key': 'GottaBeSetTo',
                                                                                                                                   'operator': 'In',
                                                                                                                                   'values': ['This']},
                                                                                                                                  {'key': 'CannotBe',
                                                                                                                                   'operator': 'NotIn',
                                                                                                                                   'values': ['ABadThing']}],
                                                                                                            'match_fields': None}]}},
         'pod_affinity': None,
         'pod_anti_affinity': None}
        """
            ).strip(),
            str(spec.affinity),
        )
        self.assertEqual(str(spec.tolerations), "None")


@needs_aws_batch
@needs_fetchable_appliance
class AWSBatchBatchSystemTest(hidden.AbstractBatchSystemTest):
    """
    Tests against the AWS Batch batch system
    """

    def supportsWallTime(self) -> bool:
        return True

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.awsBatch import AWSBatchBatchSystem

        return AWSBatchBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1e9, maxDisk=2001
        )

    def get_max_startup_seconds(self) -> int:
        # AWS Batch may need to scale out the compute environment.
        return 300


@slow
@needs_mesos
class MesosBatchSystemTest(hidden.AbstractBatchSystemTest, MesosTestSupport):
    """
    Tests against the Mesos batch system
    """

    batchSystem: "MesosBatchSystem"

    @classmethod
    def createConfig(cls) -> Config:
        """
        needs to set mesos_endpoint to localhost for testing since the default is now the
        private IP address
        """
        config = super().createConfig()
        config.mesos_endpoint = "localhost:5050"
        return config

    def supportsWallTime(self) -> bool:
        return True

    def createBatchSystem(self) -> "MesosBatchSystem":
        # We know we have Mesos so we can import the batch system
        from toil.batchSystems.mesos.batchSystem import MesosBatchSystem

        self._startMesos(numCores)
        return MesosBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1e9, maxDisk=1001
        )

    def tearDown(self) -> None:
        self._stopMesos()
        super().tearDown()

    def testIgnoreNode(self) -> None:
        self.batchSystem.ignoreNode("localhost")
        jobDesc = self._mockJobDescription(
            jobName="test2",
            unitName=None,
            jobStoreID="1",
            requirements=defaultRequirements,
        )
        job = self.batchSystem.issueBatchJob("sleep 1000", jobDesc)

        issuedID = self._waitForJobsToIssue(1)
        self.assertEqual(set(issuedID), {job})

        # Wait until a job starts or we go a while without that happening
        runningJobIDs = self._waitForJobsToStart(1, tries=20)
        # Make sure job is NOT running
        self.assertEqual(set(runningJobIDs), set({}))


def write_temp_file(s: str, temp_dir: str) -> str:
    """
    Dump a string into a temp file and return its path.
    """
    fd, path = tempfile.mkstemp(dir=temp_dir)
    try:
        encoded = s.encode("utf-8")
        assert os.write(fd, encoded) == len(encoded)
    except:
        os.unlink(path)
        raise
    else:
        return path
    finally:
        os.close(fd)


class SingleMachineBatchSystemTest(hidden.AbstractBatchSystemTest):
    """
    Tests against the single-machine batch system
    """

    def supportsWallTime(self) -> bool:
        return True

    def createBatchSystem(self) -> AbstractBatchSystem:
        return SingleMachineBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1e9, maxDisk=2001
        )

    def testProcessEscape(self, hide: bool = False) -> None:
        """
        Test to make sure that child processes and their descendants go away
        when the Toil workflow stops.

        If hide is true, will try and hide the child processes to make them
        hard to stop.
        """

        def script() -> None:
            #!/usr/bin/env python3
            import fcntl
            import os
            import signal
            import sys
            import time
            from typing import Any, Iterable

            def handle_signal(sig: Any, frame: Any) -> None:
                sys.stderr.write(f"{os.getpid()} ignoring signal {sig}\n")

            if hasattr(signal, "valid_signals"):
                # We can just ask about the signals
                all_signals: Iterable[signal.Signals] = signal.valid_signals()
            else:
                # Fish them out by name
                all_signals = [
                    getattr(signal, n)
                    for n in dir(signal)
                    if n.startswith("SIG") and not n.startswith("SIG_")
                ]

            for sig in all_signals:
                # Set up to ignore all signals we can and generally be obstinate
                if sig != signal.SIGKILL and sig != signal.SIGSTOP:
                    signal.signal(sig, handle_signal)

            if len(sys.argv) > 2:
                # Instructed to hide
                if os.fork():
                    # Try and hide the first process immediately so getting its
                    # pgid won't work.
                    sys.exit(0)

            for depth in range(3):
                # Bush out into a tree of processes
                os.fork()

            if len(sys.argv) > 1:
                fd = os.open(sys.argv[1], os.O_RDONLY)
                fcntl.lockf(fd, fcntl.LOCK_SH)

            sys.stderr.write(f"{os.getpid()} waiting...\n")

            while True:
                # Wait around forever
                time.sleep(60)

        # Get a directory where we can safely dump files.
        temp_dir = self._createTempDir()

        script_path = write_temp_file(self._getScriptSource(script), temp_dir)

        # We will have all the job processes try and lock this file shared while they are alive.
        lockable_path = write_temp_file("", temp_dir)

        try:
            command = f"{sys.executable} {script_path} {lockable_path}"
            if hide:
                # Tell the children to stop the first child and hide out in the
                # process group it made.
                command += " hide"

            # Start the job
            self.batchSystem.issueBatchJob(
                command,
                self._mockJobDescription(
                    jobName="fork", jobStoreID="1", requirements=defaultRequirements
                ),
            )
            # Wait
            time.sleep(10)

            lockfile = open(lockable_path, "w")

            if not hide:
                # In hiding mode the job will finish, and the batch system will
                # clean up after it promptly. In non-hiding mode the job will
                # stick around until shutdown, so make sure we can see it.

                # Try to lock the file and make sure it fails

                try:
                    fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
                    assert False, "Should not be able to lock file while job is running"
                except OSError:
                    pass

            # Shut down the batch system
            self.batchSystem.shutdown()

            # After the batch system shuts down, we should be able to get the
            # lock immediately, because all the children should be gone.
            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            # Then we can release it
            fcntl.lockf(lockfile, fcntl.LOCK_UN)
        finally:
            os.unlink(script_path)
            os.unlink(lockable_path)

    def testHidingProcessEscape(self) -> None:
        """
        Test to make sure that child processes and their descendants go away
        when the Toil workflow stops, even if the job process stops and leaves children.
        """

        self.testProcessEscape(hide=True)


@slow
class MaxCoresSingleMachineBatchSystemTest(ToilTest):
    """
    This test ensures that single machine batch system doesn't exceed the configured number
    cores
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        logging.basicConfig(level=logging.DEBUG)

    def setUp(self) -> None:
        super().setUp()

        temp_dir = self._createTempDir()

        # Write initial value of counter file containing a tuple of two integers (i, n) where i
        # is the number of currently executing tasks and n the maximum observed value of i
        self.counterPath = write_temp_file("0,0", temp_dir)

        def script() -> None:
            import fcntl
            import os
            import sys
            import time

            def count(delta: int) -> None:
                """
                Adjust the first integer value in a file by the given amount. If the result
                exceeds the second integer value, set the second one to the first.
                """
                fd = os.open(sys.argv[1], os.O_RDWR)
                try:
                    fcntl.flock(fd, fcntl.LOCK_EX)
                    try:
                        s = os.read(fd, 10).decode("utf-8")
                        value, maxValue = list(map(int, s.split(",")))
                        value += delta
                        if value > maxValue:
                            maxValue = value
                        os.lseek(fd, 0, 0)
                        os.ftruncate(fd, 0)
                        os.write(fd, f"{value},{maxValue}".encode())
                    finally:
                        fcntl.flock(fd, fcntl.LOCK_UN)
                finally:
                    os.close(fd)

            # Without the second argument, increment counter, sleep one second and decrement.
            # Othwerise, adjust the counter by the given delta, which can be useful for services.
            if len(sys.argv) < 3:
                count(1)
                try:
                    time.sleep(0.5)
                finally:
                    count(-1)
            else:
                count(int(sys.argv[2]))

        self.scriptPath = write_temp_file(self._getScriptSource(script), temp_dir)

    def tearDown(self) -> None:
        os.unlink(self.scriptPath)
        os.unlink(self.counterPath)

    def scriptCommand(self) -> str:
        return " ".join([sys.executable, self.scriptPath, self.counterPath])

    @retry_flaky_test(prepare=[tearDown, setUp])
    def test(self) -> None:
        # We'll use fractions to avoid rounding errors. Remember that not every fraction can be
        # represented as a floating point number.
        F = Fraction
        # This test isn't general enough to cover every possible value of minCores in
        # SingleMachineBatchSystem. Instead we hard-code a value and assert it.
        minCores = F(1, 10)
        self.assertEqual(float(minCores), SingleMachineBatchSystem.minCores)
        for maxCores in {F(minCores), minCores * 10, F(1), F(numCores, 2), F(numCores)}:
            for coresPerJob in {
                F(minCores),
                F(minCores * 10),
                F(1),
                F(maxCores, 2),
                F(maxCores),
            }:
                for load in (F(1, 10), F(1), F(10)):
                    jobs = int(maxCores / coresPerJob * load)
                    if jobs >= 1 and minCores <= coresPerJob < maxCores:
                        self.assertEqual(maxCores, float(maxCores))
                        bs = SingleMachineBatchSystem(
                            config=hidden.AbstractBatchSystemTest.createConfig(),
                            maxCores=float(maxCores),
                            # Ensure that memory or disk requirements don't get in the way.
                            maxMemory=jobs * 10,
                            maxDisk=jobs * 10,
                        )
                        try:
                            jobIds = set()
                            for i in range(0, int(jobs)):
                                desc = JobDescription(
                                    requirements=dict(
                                        cores=float(coresPerJob),
                                        memory=1,
                                        disk=1,
                                        accelerators=[],
                                        preemptible=preemptible,
                                    ),
                                    jobName=str(i),
                                    unitName="",
                                )
                                jobIds.add(bs.issueBatchJob(self.scriptCommand(), desc))
                            self.assertEqual(len(jobIds), jobs)
                            while jobIds:
                                job = bs.getUpdatedBatchJob(maxWait=10)
                                assert job is not None
                                jobId, status, wallTime = (
                                    job.jobID,
                                    job.exitStatus,
                                    job.wallTime,
                                )
                                self.assertEqual(status, 0)
                                # would raise KeyError on absence
                                jobIds.remove(jobId)
                        finally:
                            bs.shutdown()
                        concurrentTasks, maxConcurrentTasks = getCounters(
                            self.counterPath
                        )
                        self.assertEqual(concurrentTasks, 0)
                        logger.info(
                            f"maxCores: {maxCores}, "
                            f"coresPerJob: {coresPerJob}, "
                            f"load: {load}"
                        )
                        # This is the key assertion: we shouldn't run too many jobs.
                        # Because of nondeterminism we can't guarantee hitting the limit.
                        expectedMaxConcurrentTasks = min(maxCores // coresPerJob, jobs)
                        self.assertLessEqual(
                            maxConcurrentTasks, expectedMaxConcurrentTasks
                        )
                        resetCounters(self.counterPath)

    @skipIf(
        SingleMachineBatchSystem.numCores < 3,
        "Need at least three cores to run this test",
    )
    def testServices(self) -> None:
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "DEBUG"
        options.maxCores = 3
        self.assertTrue(options.maxCores <= SingleMachineBatchSystem.numCores)
        Job.Runner.startToil(Job.wrapJobFn(parentJob, self.scriptCommand()), options)
        with open(self.counterPath, "r+") as f:
            s = f.read()
        logger.info("Counter is %s", s)
        self.assertEqual(getCounters(self.counterPath), (0, 3))


# Toil can use only top-level functions so we have to add them here:


def parentJob(job: Job, cmd: str) -> None:
    job.addChildJobFn(childJob, cmd)


def childJob(job: Job, cmd: str) -> None:
    job.addService(Service(cmd))
    job.addChildJobFn(grandChildJob, cmd)
    subprocess.check_call(cmd, shell=True)


def grandChildJob(job: Job, cmd: str) -> None:
    job.addService(Service(cmd))
    job.addChildFn(greatGrandChild, cmd)
    subprocess.check_call(cmd, shell=True)


def greatGrandChild(cmd: str) -> None:
    subprocess.check_call(cmd, shell=True)


class Service(Job.Service):
    def __init__(self, cmd: str) -> None:
        super().__init__()
        self.cmd = cmd

    def start(self, job: ServiceHostJob) -> None:
        subprocess.check_call(self.cmd + " 1", shell=True)

    def check(self) -> bool:
        return True

    def stop(self, job: ServiceHostJob) -> None:
        subprocess.check_call(self.cmd + " -1", shell=True)


@slow
@needs_gridengine
class GridEngineBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
    """
    Tests against the GridEngine batch system
    """

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.gridengine import GridEngineBatchSystem

        return GridEngineBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1000e9, maxDisk=1e9
        )

    def tearDown(self) -> None:
        super().tearDown()
        # Cleanup GridEngine output log file from qsub
        from glob import glob

        for f in glob("toil_job*.o*"):
            os.unlink(f)


@slow
@needs_slurm
class SlurmBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
    """
    Tests against the Slurm batch system
    """

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.slurm import SlurmBatchSystem

        return SlurmBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1000e9, maxDisk=1e9
        )

    def tearDown(self) -> None:
        super().tearDown()
        # Cleanup 'slurm-%j.out' produced by sbatch
        from glob import glob

        for f in glob("slurm-*.out"):
            os.unlink(f)


@slow
@needs_lsf
class LSFBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
    """
    Tests against the LSF batch system
    """

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.lsf import LSFBatchSystem

        return LSFBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1000e9, maxDisk=1e9
        )


@slow
@needs_torque
class TorqueBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
    """
    Tests against the Torque batch system
    """

    def _createDummyConfig(self) -> Config:
        config = super()._createConfig()
        # can't use _getTestJobStorePath since that method removes the directory
        config.jobStore = self._createTempDir("jobStore")
        return config

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.torque import TorqueBatchSystem

        return TorqueBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1000e9, maxDisk=1e9
        )

    def tearDown(self) -> None:
        super().tearDown()
        # Cleanup 'toil_job-%j.out' produced by sbatch
        from glob import glob

        for f in glob("toil_job_*.[oe]*"):
            os.unlink(f)


@slow
@needs_htcondor
class HTCondorBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
    """
    Tests against the HTCondor batch system
    """

    def createBatchSystem(self) -> AbstractBatchSystem:
        from toil.batchSystems.htcondor import HTCondorBatchSystem

        return HTCondorBatchSystem(
            config=self.config, maxCores=numCores, maxMemory=1000e9, maxDisk=1e9
        )

    def tearDown(self) -> None:
        super().tearDown()


class TestSingleMachineBatchSystemJob(AbstractBatchSystemJobTest):
    """
    Tests Toil workflow against the SingleMachine batch system
    """

    def getBatchSystemName(self) -> str:
        return "single_machine"

    @pslow
    @pytest.mark.slow
    @retry_flaky_test(prepare=[])
    def testConcurrencyWithDisk(self, tmp_path: Path) -> None:
        """
        Tests that the batch system is allocating disk resources properly
        """

        workdir = tmp_path / "workdir"
        workdir.mkdir()
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.workDir = str(workdir)
        from toil import physicalDisk

        availableDisk = physicalDisk(options.workDir)
        logger.info("Testing disk concurrency limits with %s disk space", availableDisk)
        # More disk might become available by the time Toil starts, so we limit it here
        options.maxDisk = availableDisk
        options.batchSystem = self.getBatchSystemName()

        counterPath = tmp_path / "counter"
        resetCounters(counterPath)
        value, maxValue = getCounters(counterPath)
        assert (value, maxValue) == (0, 0)

        half_disk = availableDisk // 2
        more_than_half_disk = half_disk + 500
        logger.info("Dividing into parts of %s and %s", half_disk, more_than_half_disk)

        root = Job()
        # Physically, we're asking for 50% of disk and 50% of disk + 500bytes in the two jobs. The
        # batchsystem should not allow the 2 child jobs to run concurrently.
        root.addChild(
            Job.wrapFn(
                measureConcurrency,
                counterPath,
                self.sleepTime,
                cores=1,
                memory="1M",
                disk=half_disk,
            )
        )
        root.addChild(
            Job.wrapFn(
                measureConcurrency,
                counterPath,
                self.sleepTime,
                cores=1,
                memory="1M",
                disk=more_than_half_disk,
            )
        )
        Job.Runner.startToil(root, options)
        _, maxValue = getCounters(counterPath)

        logger.info("After run: %s disk space", physicalDisk(options.workDir))

        assert maxValue == 1

    @pytest.mark.skipif(
        SingleMachineBatchSystem.numCores < 4,
        reason="Need at least four cores to run this test",
    )
    @pslow
    @pytest.mark.slow
    def testNestedResourcesDoNotBlock(self, tmp_path: Path) -> None:
        """
        Resources are requested in the order Memory > Cpu > Disk.
        Test that unavailability of cpus for one job that is scheduled does not block another job
        that can run.
        """
        workdir = tmp_path / "workdir"
        workdir.mkdir()
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.workDir = str(workdir)
        options.maxCores = 4
        from toil import physicalMemory

        availableMemory = physicalMemory()
        options.batchSystem = self.getBatchSystemName()

        outFile = tmp_path / "counter"
        outFile.open("w").close()

        root = Job()

        blocker = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=30,
            writeVal="b",
            cores=2,
            memory="1M",
            disk="1M",
        )
        firstJob = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=5,
            writeVal="fJ",
            cores=1,
            memory="1M",
            disk="1M",
        )
        secondJob = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=10,
            writeVal="sJ",
            cores=1,
            memory="1M",
            disk="1M",
        )

        # Should block off 50% of memory while waiting for it's 3 cores
        firstJobChild = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=0,
            writeVal="fJC",
            cores=3,
            memory=int(availableMemory // 2),
            disk="1M",
        )

        # These two shouldn't be able to run before B because there should be only
        # (50% of memory - 1M) available (firstJobChild should be blocking 50%)
        secondJobChild = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=5,
            writeVal="sJC",
            cores=2,
            memory=int(availableMemory // 1.5),
            disk="1M",
        )
        secondJobGrandChild = Job.wrapFn(
            _resourceBlockTestAuxFn,
            outFile=outFile,
            sleepTime=5,
            writeVal="sJGC",
            cores=2,
            memory=int(availableMemory // 1.5),
            disk="1M",
        )

        root.addChild(blocker)
        root.addChild(firstJob)
        root.addChild(secondJob)

        firstJob.addChild(firstJobChild)
        secondJob.addChild(secondJobChild)

        secondJobChild.addChild(secondJobGrandChild)
        """
        The tree is:
                    root
                  /   |   \
                 b    fJ   sJ
                      |    |
                      fJC  sJC
                           |
                           sJGC
        But the order of execution should be
        root > b , fJ, sJ > sJC > sJGC > fJC
        since fJC cannot run till bl finishes but sJC and sJGC can(fJC blocked by disk). If the
        resource acquisition is written properly, then fJC which is scheduled before sJC and sJGC
        should not block them, and should only run after they finish.
        """
        Job.Runner.startToil(root, options)
        with outFile.open() as oFH:
            outString = oFH.read()
        # The ordering of b, fJ and sJ is non-deterministic since they are scheduled at the same
        # time. We look for all possible permutations.
        possibleStarts = tuple(
            "".join(x) for x in itertools.permutations(["b", "fJ", "sJ"])
        )
        assert outString.startswith(possibleStarts)
        assert outString.endswith("sJCsJGCfJC")


def _resourceBlockTestAuxFn(outFile: StrPath, sleepTime: int, writeVal: str) -> None:
    """
    Write a value to the out file and then sleep for requested seconds.
    :param outFile: File to write to
    :param sleepTime: Time to sleep for
    :param writeVal: Character to write
    """
    with open(outFile, "a") as oFH:
        fcntl.flock(oFH, fcntl.LOCK_EX)
        oFH.write(writeVal)
    time.sleep(sleepTime)


@pslow
@pytest.mark.slow
@pneeds_mesos
class TestMesosBatchSystemJob(AbstractBatchSystemJobTest, MesosTestSupport):
    """
    Tests Toil workflow against the Mesos batch system
    """

    @pytest.fixture(autouse=True)
    def mesos_support(self) -> Generator[None]:
        try:
            self._startMesos(self.cpuCount)
            yield
        finally:
            self._stopMesos()

    def getOptions(self, tempDir: Path) -> argparse.Namespace:
        options = super().getOptions(tempDir)
        options.mesos_endpoint = "localhost:5050"
        return options

    def getBatchSystemName(self) -> "str":
        return "mesos"


def measureConcurrency(filepath: StrPath, sleep_time: int = 10) -> int:
    """
    Run in parallel to determine the number of concurrent tasks.
    This code was copied from toil.batchSystemTestMaxCoresSingleMachineBatchSystemTest
    :param filepath: path to counter file
    :param sleep_time: number of seconds to sleep before counting down
    :return: max concurrency value
    """
    count(1, filepath)
    try:
        time.sleep(sleep_time)
    finally:
        return count(-1, filepath)


def count(delta: int, file_path: StrPath) -> int:
    """
    Increments counter file and returns the max number of times the file
    has been modified. Counter data must be in the form:
    concurrent tasks, max concurrent tasks (counter should be initialized to 0,0)

    :param delta: increment value
    :param file_path: path to shared counter file
    :return: max concurrent tasks
    """

    fd = os.open(file_path, os.O_RDWR)
    try:
        fcntl.flock(fd, fcntl.LOCK_EX)
        try:
            s = os.read(fd, 10)
            value, maxValue = (int(i) for i in s.decode("utf-8").split(","))
            value += delta
            if value > maxValue:
                maxValue = value
            os.lseek(fd, 0, 0)
            os.ftruncate(fd, 0)
            os.write(fd, f"{value},{maxValue}".encode())
        finally:
            fcntl.flock(fd, fcntl.LOCK_UN)
    finally:
        os.close(fd)
    return maxValue


def getCounters(path: StrPath) -> tuple[int, int]:
    with open(path, "r+") as f:
        concurrentTasks, maxConcurrentTasks = (int(i) for i in f.read().split(","))
    return concurrentTasks, maxConcurrentTasks


def resetCounters(path: StrPath) -> None:
    with open(path, "w") as f:
        f.write("0,0")
        f.close()


def get_omp_threads() -> str:
    return os.environ["OMP_NUM_THREADS"]



from types import ModuleType
from typing import Any, Callable, Sequence, Tuple

from _typeshed import Incomplete

class AUTO_TOPIC: ...

UserListener = Callable[..., Any]

def getModule(obj: Any) -> ModuleType: ...
def getID(callable_obj: UserListener) -> Tuple[str, ModuleType]: ...
def getRawFunction(callable_obj: UserListener) -> Tuple[Callable[..., Any]]: ...

class ListenerMismatchError(ValueError):
    args: Incomplete
    msg: Incomplete
    module: Incomplete
    idStr: Incomplete
    def __init__(self, msg: str, listener: UserListener, *args: Any) -> None: ...

class CallArgsInfo:
    autoTopicArgName: Incomplete
    acceptsAllKwargs: bool
    requiredArgs: Incomplete
    optionalArgs: Incomplete
    allParams: Incomplete
    def __init__(self, func: UserListener, ignoreArgs: Sequence[str] = ...) -> None: ...
    def getAllArgs(self) -> Tuple[str]: ...
    def getOptionalArgs(self) -> Tuple[str]: ...
    def getRequiredArgs(self) -> Tuple[str]: ...

def getArgs(
    callable_obj: UserListener, ignoreArgs: Sequence[str] = ...
) -> CallArgsInfo: ...



# Copyright (C) 2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains functions for integrating Toil with UCSC Dockstore, for reporting metrics.

For basic TRS functionality for fetching workflows, see trs.py.
"""

import datetime
import logging
import math
import os
import re
import sys
import uuid
from typing import Any, Literal, Optional, Union, TypedDict, cast

from urllib.parse import urlparse, unquote, quote
import requests

from toil.lib.misc import unix_seconds_to_timestamp, seconds_to_duration
from toil.lib.trs import TRS_ROOT
from toil.lib.retry import retry
from toil.lib.web import web_session
from toil.version import baseVersion

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

logger = logging.getLogger(__name__)

# We assume TRS_ROOT is actually a Dockstore instance.

# This is a publish-able token for production Dockstore for Toil to use.
# This is NOT a secret value.
DEFAULT_DOCKSTORE_TOKEN = "2bff46294daddef6df185452b04db6143ea8a59f52ee3c325d3e1df418511b7d"

# How should we authenticate our Dockstore requests?
DOCKSTORE_TOKEN = os.environ.get("TOIL_DOCKSTORE_TOKEN", DEFAULT_DOCKSTORE_TOKEN)

# What platform should we report metrics as?
DOCKSTORE_PLATFORM = "TOIL"


# This is a https://schema.org/CompletedActionStatus
# The values here are from expanding the type info in the Docksotre docs at
# <https://dockstore.org/api/static/swagger-ui/index.html#/extendedGA4GH/executionMetricsPost>
ExecutionStatus = Union[Literal["ALL"], Literal["SUCCESSFUL"], Literal["FAILED"], Literal["FAILED_SEMANTIC_INVALID"], Literal["FAILED_RUNTIME_INVALID"], Literal["ABORTED"]]

class Cost(TypedDict):
    """
    Representation of the cost of running something.
    """

    value: float
    """
    Cost in US Dollars.
    """

class RunExecution(TypedDict):
    """
    Dockstore metrics data for a workflow or task run.
    """

    executionId: str
    """
    Executor-generated unique execution ID for this workflow or task.
    """

    dateExecuted: str
    """
    ISO 8601 UTC timestamp when the execution happend.
    """

    executionStatus: ExecutionStatus
    """
    Did the execution work?
    """

    executionTime: NotRequired[str]
    """
    Total time of the run in ISO 8601 duration format.
    """

    # TODO: Is this meant to be actual usage or amount provided?
    memoryRequirementsGB: NotRequired[float]
    """
    Memory required for the execution in gigabytes (not GiB).
    """

    cpuRequirements: NotRequired[int]
    """
    Number of CPUs required.
    """

    cost: NotRequired[Cost]
    """
    How much the execution cost to run.
    """

    # TODO: What if two cloud providers have the same region naming scheme?
    region: NotRequired[str]
    """
    The (cloud) region the workflow was executed in.
    """

    additionalProperties: NotRequired[dict[str, Any]]
    """
    Any additional properties to send.

    Dockstore can take any JSON-able structured data.
    """

class TaskExecutions(TypedDict):
    """
    Dockstore metrics data for all the tasks in a workflow.
    """

    # TODO: Right now we use different IDs for the workflow RunExecution and
    # for its corresponding collection of TaskExecutions, so there's no nice
    # way to find the one from the other.
    executionId: str
    """
    Executor-generated unique execution ID.
    """

    dateExecuted: str
    """
    ISO 8601 UTC timestamp when the execution happend.
    """

    taskExecutions: list[RunExecution]
    """
    Individual executions of each task in the workflow.
    """

    additionalProperties: NotRequired[dict[str, Any]]
    """
    Any additional properties to send.

    Dockstore can take any JSON-able structured data.
    """

def ensure_valid_id(execution_id: str) -> None:
    """
    Make sure the given execution ID is in Dockstore format and will be accepted by Dockstore.

    Must be alphanumeric (with internal underscores allowed) and <100
    characters long.

    :raises ValueError: if the ID is not in the right format
    """
    if len(execution_id) >= 100:
        raise ValueError("Execution ID too long")
    if len(execution_id) == 0:
        raise ValueError("Execution ID must not be empty")
    if execution_id[0] == "_" or execution_id[-1] == "_":
        raise ValueError("Execution ID must not start or end with an underscore")
    if not re.fullmatch("[a-zA-Z0-9_]+", execution_id):
        raise ValueError("Execution ID must be alphanumeric with internal underscores")

def pack_workflow_metrics(
        execution_id: str,
        start_time: float,
        runtime: float,
        succeeded: bool,
        job_store_type: Optional[str] = None,
        batch_system: Optional[str] = None,
        caching: Optional[bool] = None,
        toil_version: Optional[str] = None,
        python_version: Optional[str] = None,
        platform_system: Optional[str] = None,
        platform_machine: Optional[str] = None
    ) -> RunExecution:
    """
    Pack up per-workflow metrics into a format that can be submitted to Dockstore.

    :param execution_id: Unique ID for the workflow execution. Must be in
        Dockstore format. 
    :param start_time: Execution start time in seconds since the Unix epoch.
    :param rutime: Execution duration in seconds.
    :param jobstore_type: Kind of job store used, like "file" or "aws".
    :param batch_system: Python class name implementing the batch system used.
    :param caching: Whether Toil filestore-level cahcing was used.
    :param toil_version: Version of Toil used (without any Git hash).
    :param python_version: Version of Python used.
    :param platform_system: Operating system type (like "Darwin" or "Linux").
    :param platform_machine: Machine architecture of the leader (like "AMD64").
    """

    # Enforce Dockstore's constraints
    ensure_valid_id(execution_id)

    # Pack up into a RunExecution
    result = RunExecution(
        executionId=execution_id,
        dateExecuted=unix_seconds_to_timestamp(start_time),
        executionTime=seconds_to_duration(runtime),
        executionStatus="SUCCESSFUL" if succeeded else "FAILED"
    )

    # TODO: Just use kwargs here?
    additional_properties: dict[str, Any] = {}

    if job_store_type is not None:
        additional_properties["jobStoreType"] = job_store_type

    if batch_system is not None:
        additional_properties["batchSystem"] = batch_system

    if caching is not None:
        additional_properties["caching"] = caching

    if toil_version is not None:
        additional_properties["toilVersion"] = toil_version

    if python_version is not None:
        additional_properties["pythonVersion"] = python_version

    if platform_system is not None:
        additional_properties["platformSystem"] = platform_system

    if platform_machine is not None:
        additional_properties["platformMachine"] = platform_machine

    if len(additional_properties) > 0:
        result["additionalProperties"] = additional_properties

    return result

def pack_single_task_metrics(
        execution_id: str,
        start_time: float,
        runtime: float,
        succeeded: bool,
        job_name: Optional[str] = None,
        cores: Optional[float] = None,
        cpu_seconds: Optional[float] = None,
        memory_bytes: Optional[int] = None,
        disk_bytes: Optional[int] = None,
    ) -> RunExecution:
    """
    Pack up metrics for a single task execution in a format that can be used in a Dockstore submission.

    :param execution_id: Unique ID for the workflow execution. Must be in
        Dockstore format. 
    :param start_time: Execution start time in seconds since the Unix epoch.
    :param rutime: Execution duration in seconds.
    :param succeeded: Whether the execution succeeded.
    :param job_name: Name of the job within the workflow.
    :param cores: CPU cores allocated to the job.
    :param cpu_seconds: CPU seconds consumed by the job.
    :param memory_bytes: Memory consumed by the job in bytes.
    :param disk_bytes: Disk space consumed by the job in bytes.
    """

    # TODO: Deduplicate with workflow code since the output type is the same.

    # Enforce Dockstore's constraints
    ensure_valid_id(execution_id)

    # Pack up into a RunExecution
    result = RunExecution(
        executionId=execution_id,
        dateExecuted=unix_seconds_to_timestamp(start_time),
        executionTime=seconds_to_duration(runtime),
        executionStatus="SUCCESSFUL" if succeeded else "FAILED"
    )

    if memory_bytes is not None:
        # Convert bytes to fractional gigabytes
        result["memoryRequirementsGB"] = memory_bytes / 1_000_000_000

    if cores is not None:
        # Convert possibly fractional cores to an integer for Dockstore
        result["cpuRequirements"] = int(math.ceil(cores))

    # TODO: Just use kwargs here?
    additional_properties: dict[str, Any] = {}

    if job_name is not None:
        # Convert to Doskstore-style camelCase property keys
        additional_properties["jobName"] = job_name

    if disk_bytes is not None:
        # Convert to a Dockstore-style fractional disk gigabytes
        additional_properties["diskRequirementsGB"] = disk_bytes / 1_000_000_000

    if cpu_seconds is not None:
        # Use a Dockstore-ier name here too
        additional_properties["cpuRequirementsCoreSeconds"] = cpu_seconds

    if len(additional_properties) > 0:
        result["additionalProperties"] = additional_properties

    return result


def pack_workflow_task_set_metrics(execution_id: str, start_time: float, tasks: list[RunExecution]) -> TaskExecutions:
    """
    Pack up metrics for all the tasks in a workflow execution into a format that can be submitted to Dockstore.
    
    :param execution_id: Unique ID for the workflow execution. Must be in
        Dockstore format. 
    :param start_time: Execution start time for the overall workflow execution
        in seconds since the Unix epoch.
    :param tasks: Packed tasks from pack_single_task_metrics()
    """

    # Enforce Dockstore's constraints
    ensure_valid_id(execution_id)

    return TaskExecutions(
        executionId=execution_id,
        dateExecuted=unix_seconds_to_timestamp(start_time),
        taskExecutions=tasks
    )

def send_metrics(trs_workflow_id: str, trs_version: str, workflow_runs: list[RunExecution], workflow_task_sets: list[TaskExecutions]) -> None:
    """
    Send packed workflow and/or task metrics to Dockstore.

    :param workflow_runs: list of packed metrics objects for each workflow.

    :param workflow_task_sets: list of packed metrics objects for the tasks in
        each workflow. Each workflow should have one entry containing all its
        tasks. Does not have to be the same order/set of workflows as
        workflow_runs.
    
    :raises requests.HTTPError: if Dockstore does not accept the metrics.
    """

    # Aggregate into a submission
    to_post = {
        "runExecutions": workflow_runs,
        "taskExecutions": workflow_task_sets,
        "validationExecutions": []
    }

    # Set the submission query string metadata
    submission_params = {
        "platform": DOCKSTORE_PLATFORM,
        "description": "Workflow status from Toil"
    }

    # Set the headers. Even though user agent isn't in here, it still gets
    # sent.
    headers = {}
    if DOCKSTORE_TOKEN is not None:
        headers["Authorization"] = f"Bearer {DOCKSTORE_TOKEN}"

    # Note that Dockstore's metrics apparently need two levels of /api for some reason.
    endpoint_url = f"{TRS_ROOT}/api/api/ga4gh/v2/extended/{quote(trs_workflow_id, safe='')}/versions/{quote(trs_version, safe='')}/executions"

    logger.info("Sending workflow metrics to %s", endpoint_url)
    logger.debug("With data: %s", to_post)
    logger.debug("With headers: %s", headers)

    try:
        result = web_session.post(endpoint_url, params=submission_params, json=to_post, headers=headers)
        result.raise_for_status()
        logger.debug("Workflow metrics were accepted by Dockstore. Dockstore response code: %s", result.status_code)
    except requests.HTTPError as e:
        logger.warning("Workflow metrics were not accepted by Dockstore. Dockstore complained: %s", e.response.text)
        raise

def get_metrics_url(trs_workflow_id: str, trs_version: str, execution_id: str) -> str:
    """
    Get the URL where a workflow metrics object (for a workflow, or for a set of tasks) can be fetched back from.
    """

    return f"{TRS_ROOT}/api/api/ga4gh/v2/extended/{quote(trs_workflow_id, safe='')}/versions/{quote(trs_version, safe='')}/execution?platform={DOCKSTORE_PLATFORM}&executionId={quote(execution_id, safe='')}"




# Copyright (C) 2015-2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging
import pytest
from typing import IO
import urllib.request
from urllib.error import URLError

from toil.lib.retry import retry
from toil.lib.trs import find_workflow, fetch_workflow
from toil.test import ToilTest, needs_online

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

@pytest.mark.integrative
@needs_online
class DockstoreLookupTest(ToilTest):
    """
    Make sure we can look up workflows on Dockstore.
    """

    @retry(errors=[URLError, RuntimeError])
    def read_result(self, url_or_path: str) -> IO[bytes]:
        """
        Read a file or URL.

        Binary mode to allow testing for binary file support.

        This lets us test that we have the right workflow contents and not care
        how we are being shown them.
        """
        if url_or_path.startswith("http://") or url_or_path.startswith("https://"):
            response = urllib.request.urlopen(url_or_path)
            if response.status != 200:
                raise RuntimeError(f"HTTP error response: {response}")
            return response
        else:
            return open(url_or_path, "rb")

    # TODO: Tests that definitely test a clear cache

    def test_lookup_from_page_url(self) -> None:
        PAGE_URL = "https://dockstore.org/workflows/github.com/dockstore/bcc2020-training/HelloWorld:master?tab=info"
        trs_id, trs_version, language = find_workflow(PAGE_URL)

        self.assertEqual(trs_id, "#workflow/github.com/dockstore/bcc2020-training/HelloWorld")
        self.assertEqual(trs_version, "master")
        self.assertEqual(language, "WDL")

    def test_lookup_from_trs_with_version(self) -> None:
        TRS_ID = "#workflow/github.com/dockstore-testing/md5sum-checker"
        TRS_VERSION = "master"
        trs_id, trs_version, language = find_workflow(f"{TRS_ID}:{TRS_VERSION}")
        
        self.assertEqual(trs_id, TRS_ID)
        self.assertEqual(trs_version, TRS_VERSION)
        self.assertEqual(language, "CWL")

    def test_lookup_from_trs_no_version(self) -> None:
        TRS_ID = "#workflow/github.com/dockstore-testing/md5sum-checker"
        with pytest.raises(ValueError):
            # We don't yet have a way to read Dockstore's default version info,
            # so it's not safe to apply any default version when multiple
            # versions exist.
            trs_id, trs_version, language = find_workflow(TRS_ID)

    # TODO: Add a test with a workflow that we know has and will only ever
    # have one version, to test version auto-detection in that case.

    def test_get(self) -> None:
        TRS_ID = "#workflow/github.com/dockstore-testing/md5sum-checker"
        TRS_VERSION = "master"
        LANGUAGE = "CWL"
        # Despite "-checker" in the ID, this actually refers to the base md5sum
        # workflow that just happens to have a checker *available*, not to the
        # checker workflow itself.
        WORKFLOW_URL = "https://raw.githubusercontent.com/dockstore-testing/md5sum-checker/master/md5sum/md5sum-workflow.cwl"
        looked_up = fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)
        
        data_from_lookup = self.read_result(looked_up).read()
        data_from_source = self.read_result(WORKFLOW_URL).read()
        self.assertEqual(data_from_lookup, data_from_source)

    def test_get_from_trs_cached(self) -> None:
        TRS_ID = "#workflow/github.com/dockstore-testing/md5sum-checker"
        TRS_VERSION = "master"
        LANGUAGE = "CWL"
        WORKFLOW_URL = "https://raw.githubusercontent.com/dockstore-testing/md5sum-checker/master/md5sum/md5sum-workflow.cwl"
        # This lookup may or may not be cached
        fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)
        # This lookup is definitely cached
        looked_up = fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)
        
        data_from_lookup = self.read_result(looked_up).read()
        data_from_source = self.read_result(WORKFLOW_URL).read()
        self.assertEqual(data_from_lookup, data_from_source)

    def test_lookup_from_trs_with_version(self) -> None:
        TRS_VERSIONED_ID = "#workflow/github.com/dockstore-testing/md5sum-checker:workflowWithHTTPImport"
        trs_id, trs_version, language = find_workflow(TRS_VERSIONED_ID)
        
        parts = TRS_VERSIONED_ID.split(":")

        self.assertEqual(trs_id, parts[0])
        self.assertEqual(trs_version, parts[1])
        self.assertEqual(language, "CWL")

    def test_lookup_from_trs_nonexistent_workflow(self) -> None:
        TRS_VERSIONED_ID = "#workflow/github.com/adamnovak/veryfakerepo:notARealVersion"
        with self.assertRaises(FileNotFoundError):
            looked_up = find_workflow(TRS_VERSIONED_ID)

    def test_lookup_from_trs_nonexistent_workflow_bad_format(self) -> None:
        TRS_VERSIONED_ID = "#workflow/AbsoluteGarbage:notARealVersion"
        with self.assertRaises(FileNotFoundError):
            looked_up = find_workflow(TRS_VERSIONED_ID)

    def test_lookup_from_trs_nonexistent_version(self) -> None:
        TRS_VERSIONED_ID = "#workflow/github.com/dockstore-testing/md5sum-checker:notARealVersion"
        with self.assertRaises(FileNotFoundError):
            looked_up = find_workflow(TRS_VERSIONED_ID)
    
    def test_get_nonexistent_workflow(self) -> None:
        TRS_ID = "#workflow/github.com/adamnovak/veryfakerepo"
        TRS_VERSION = "notARealVersion"
        LANGUAGE = "CWL"
        with self.assertRaises(FileNotFoundError):
            looked_up = fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)

    def test_get_nonexistent_version(self) -> None:
        TRS_ID = "#workflow/github.com/dockstore-testing/md5sum-checker"
        TRS_VERSION = "notARealVersion"
        LANGUAGE = "CWL"
        with self.assertRaises(FileNotFoundError):
            looked_up = fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)

    def test_get_nonexistent_workflow_bad_format(self) -> None:
        # Dockstore enforces an ID pattern and blames your request if you ask
        # about something that doesn't follow it. So don't follow it.
        TRS_ID = "#workflow/AbsoluteGarbage"
        TRS_VERSION = "notARealVersion"
        LANGUAGE = "CWL"
        with self.assertRaises(FileNotFoundError):
            looked_up = fetch_workflow(TRS_ID, TRS_VERSION, LANGUAGE)






import sys

from configargparse import ArgumentParser

from toil.common import Toil
from toil.job import Job


def main():
    """
    This workflow always fails.

    Invoke like:

        python examples/example_alwaysfail.py ./jobstore

    Then you can inspect the job store with tools like `toil status`:

        toil status --logs ./jobstore

    """
    parser = ArgumentParser(description=main.__doc__)
    Job.Runner.addToilOptions(parser)
    options = parser.parse_args(sys.argv[1:])

    hello_job = Job.wrapJobFn(explode)

    with Toil(options) as toil:
        toil.start(hello_job)


def explode(job):
    sys.stderr.write("Something somewhere has gone terribly wrong\n")
    raise RuntimeError("Boom!")


if __name__ == "__main__":
    main()
